# Copyright 2023 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

cimport cpython
from cython.operator cimport dereference

import enum
import functools
import logging
import os
from threading import Thread
from typing import List, Mapping, Tuple, Union

import _observability

# Time we wait for batch exporting census data
# TODO(xuanwn): change interval to a more appropriate number
CENSUS_EXPORT_BATCH_INTERVAL_SECS = float(os.environ.get('GRPC_PYTHON_CENSUS_EXPORT_BATCH_INTERVAL_SECS', 0.5))
GRPC_PYTHON_CENSUS_EXPORT_THREAD_TIMEOUT = float(os.environ.get('GRPC_PYTHON_CENSUS_EXPORT_THREAD_TIMEOUT', 10))
cdef const char* CLIENT_CALL_TRACER = "client_call_tracer"
cdef const char* SERVER_CALL_TRACER_FACTORY = "server_call_tracer_factory"
cdef bint GLOBAL_SHUTDOWN_EXPORT_THREAD = False
cdef object GLOBAL_EXPORT_THREAD

_LOGGER = logging.getLogger(__name__)


class _CyMetricsName:
  CY_CLIENT_API_LATENCY = kRpcClientApiLatencyMeasureName
  CY_CLIENT_SNET_MESSSAGES_PER_RPC = kRpcClientSentMessagesPerRpcMeasureName
  CY_CLIENT_SEND_BYTES_PER_RPC = kRpcClientSentBytesPerRpcMeasureName
  CY_CLIENT_RECEIVED_MESSAGES_PER_RPC = kRpcClientReceivedMessagesPerRpcMeasureName
  CY_CLIENT_RECEIVED_BYTES_PER_RPC = kRpcClientReceivedBytesPerRpcMeasureName
  CY_CLIENT_ROUNDTRIP_LATENCY = kRpcClientRoundtripLatencyMeasureName
  CY_CLIENT_COMPLETED_RPC = kRpcClientCompletedRpcMeasureName
  CY_CLIENT_SERVER_LATENCY = kRpcClientServerLatencyMeasureName
  CY_CLIENT_STARTED_RPCS = kRpcClientStartedRpcsMeasureName
  CY_CLIENT_RETRIES_PER_CALL = kRpcClientRetriesPerCallMeasureName
  CY_CLIENT_TRANSPARENT_RETRIES_PER_CALL = kRpcClientTransparentRetriesPerCallMeasureName
  CY_CLIENT_RETRY_DELAY_PER_CALL = kRpcClientRetryDelayPerCallMeasureName
  CY_CLIENT_TRANSPORT_LATENCY = kRpcClientTransportLatencyMeasureName
  CY_SERVER_SENT_MESSAGES_PER_RPC = kRpcServerSentMessagesPerRpcMeasureName
  CY_SERVER_SENT_BYTES_PER_RPC = kRpcServerSentBytesPerRpcMeasureName
  CY_SERVER_RECEIVED_MESSAGES_PER_RPC = kRpcServerReceivedMessagesPerRpcMeasureName
  CY_SERVER_RECEIVED_BYTES_PER_RPC = kRpcServerReceivedBytesPerRpcMeasureName
  CY_SERVER_SERVER_LATENCY = kRpcServerServerLatencyMeasureName
  CY_SERVER_COMPLETED_RPC = kRpcServerCompletedRpcMeasureName
  CY_SERVER_STARTED_RPCS = kRpcServerStartedRpcsMeasureName

@enum.unique
class MetricsName(enum.Enum):
  CLIENT_STARTED_RPCS = _CyMetricsName.CY_CLIENT_STARTED_RPCS
  CLIENT_API_LATENCY = _CyMetricsName.CY_CLIENT_API_LATENCY
  CLIENT_SNET_MESSSAGES_PER_RPC = _CyMetricsName.CY_CLIENT_SNET_MESSSAGES_PER_RPC
  CLIENT_SEND_BYTES_PER_RPC = _CyMetricsName.CY_CLIENT_SEND_BYTES_PER_RPC
  CLIENT_RECEIVED_MESSAGES_PER_RPC = _CyMetricsName.CY_CLIENT_RECEIVED_MESSAGES_PER_RPC
  CLIENT_RECEIVED_BYTES_PER_RPC = _CyMetricsName.CY_CLIENT_RECEIVED_BYTES_PER_RPC
  CLIENT_ROUNDTRIP_LATENCY = _CyMetricsName.CY_CLIENT_ROUNDTRIP_LATENCY
  CLIENT_COMPLETED_RPC = _CyMetricsName.CY_CLIENT_COMPLETED_RPC
  CLIENT_SERVER_LATENCY = _CyMetricsName.CY_CLIENT_SERVER_LATENCY
  CLIENT_RETRIES_PER_CALL = _CyMetricsName.CY_CLIENT_RETRIES_PER_CALL
  CLIENT_TRANSPARENT_RETRIES_PER_CALL = _CyMetricsName.CY_CLIENT_TRANSPARENT_RETRIES_PER_CALL
  CLIENT_RETRY_DELAY_PER_CALL = _CyMetricsName.CY_CLIENT_RETRY_DELAY_PER_CALL
  CLIENT_TRANSPORT_LATENCY = _CyMetricsName.CY_CLIENT_TRANSPORT_LATENCY
  SERVER_SENT_MESSAGES_PER_RPC = _CyMetricsName.CY_SERVER_SENT_MESSAGES_PER_RPC
  SERVER_SENT_BYTES_PER_RPC = _CyMetricsName.CY_SERVER_SENT_BYTES_PER_RPC
  SERVER_RECEIVED_MESSAGES_PER_RPC = _CyMetricsName.CY_SERVER_RECEIVED_MESSAGES_PER_RPC
  SERVER_RECEIVED_BYTES_PER_RPC = _CyMetricsName.CY_SERVER_RECEIVED_BYTES_PER_RPC
  SERVER_SERVER_LATENCY = _CyMetricsName.CY_SERVER_SERVER_LATENCY
  SERVER_COMPLETED_RPC = _CyMetricsName.CY_SERVER_COMPLETED_RPC
  SERVER_STARTED_RPCS = _CyMetricsName.CY_SERVER_STARTED_RPCS

# Delay map creation due to circular dependencies
_CY_METRICS_NAME_TO_PY_METRICS_NAME_MAPPING = {x.value: x for x in MetricsName}

def cyobservability_init(object exporter) -> None:
  exporter: _observability.Exporter

  NativeObservabilityInit()
  _start_exporting_thread(exporter)


def _start_exporting_thread(object exporter) -> None:
  exporter: _observability.Exporter

  global GLOBAL_EXPORT_THREAD
  global GLOBAL_SHUTDOWN_EXPORT_THREAD
  GLOBAL_SHUTDOWN_EXPORT_THREAD = False
  # TODO(xuanwn): Change it to daemon thread.
  GLOBAL_EXPORT_THREAD = Thread(target=_export_census_data, args=(exporter,))
  GLOBAL_EXPORT_THREAD.start()


def set_gcp_observability_config(object py_config) -> bool:
  py_config: _gcp_observability.GcpObservabilityPythonConfig

  py_labels = {}
  sampling_rate = 0.0

  cdef cGcpObservabilityConfig c_config = ReadAndActivateObservabilityConfig()
  if not c_config.is_valid:
    return False

  for label in c_config.labels:
    py_labels[_decode(label.key)] = _decode(label.value)

  if PythonCensusTracingEnabled():
    sampling_rate = c_config.cloud_trace.sampling_rate
    # Save sampling rate to global sampler.
    ProbabilitySampler.Get().SetThreshold(sampling_rate)

  py_config.set_configuration(_decode(c_config.project_id), sampling_rate, py_labels,
                              PythonCensusTracingEnabled(), PythonCensusStatsEnabled())
  return True


def create_client_call_tracer(bytes method_name, bytes trace_id,
                              bytes parent_span_id=b'') -> cpython.PyObject:
  """Create a ClientCallTracer and save to PyCapsule.

  Returns: A grpc_observability._observability.ClientCallTracerCapsule object.
  """
  cdef char* c_method = cpython.PyBytes_AsString(method_name)
  cdef char* c_trace_id = cpython.PyBytes_AsString(trace_id)
  cdef char* c_parent_span_id = cpython.PyBytes_AsString(parent_span_id)

  cdef void* call_tracer = CreateClientCallTracer(c_method, c_trace_id, c_parent_span_id)
  capsule = cpython.PyCapsule_New(call_tracer, CLIENT_CALL_TRACER, NULL)
  return capsule


def create_server_call_tracer_factory_capsule() -> cpython.PyObject:
  """Create a ServerCallTracerFactory and save to PyCapsule.

  Returns: A grpc_observability._observability.ServerCallTracerFactoryCapsule object.
  """
  cdef void* call_tracer_factory = CreateServerCallTracerFactory()
  capsule = cpython.PyCapsule_New(call_tracer_factory, SERVER_CALL_TRACER_FACTORY, NULL)
  return capsule


def delete_client_call_tracer(object client_call_tracer) -> None:
  client_call_tracer: grpc._observability.ClientCallTracerCapsule

  if cpython.PyCapsule_IsValid(client_call_tracer, CLIENT_CALL_TRACER):
    capsule_ptr = cpython.PyCapsule_GetPointer(client_call_tracer, CLIENT_CALL_TRACER)
    call_tracer_ptr = <ClientCallTracer*>capsule_ptr
    del call_tracer_ptr


def _c_label_to_labels(vector[Label] c_labels) -> Mapping[str, str]:
  py_labels = {}
  for label in c_labels:
    py_labels[_decode(label.key)] = _decode(label.value)
  return py_labels


def _c_measurement_to_measurement(object measurement
  ) -> Mapping[str, Union[enum, Mapping[str, Union[float, int]]]]:
  """Convert Cython Measurement to Python measurement.

  Args:
  measurement: Actual measurement repesented by Cython type Measurement, using object here
   since Cython refuse to automatically convert a union with unsafe type combinations.

  Returns:
   A mapping object with keys and values as following:
    name -> cMetricsName
    type -> MeasurementType
    value -> {value_double: float | value_int: int}
  """
  measurement: Measurement

  py_measurement = {}
  py_measurement['name'] = measurement['name']
  py_measurement['type'] = measurement['type']
  if measurement['type'] == kMeasurementDouble:
    py_measurement['value'] = {'value_double': measurement['value']['value_double']}
  else:
    py_measurement['value'] = {'value_int': measurement['value']['value_int']}
  return py_measurement


def _c_annotation_to_annotations(vector[Annotation] c_annotations) -> List[Tuple[str, str]]:
  py_annotations = []
  for annotation in c_annotations:
    py_annotations.append((_decode(annotation.time_stamp),
                          _decode(annotation.description)))
  return py_annotations


def observability_deinit() -> None:
  _shutdown_exporting_thread()
  EnablePythonCensusStats(False)
  EnablePythonCensusTracing(False)


@functools.lru_cache(maxsize=None)
def _cy_metric_name_to_py_metric_name(cMetricsName metric_name) -> MetricsName:
  try:
    return _CY_METRICS_NAME_TO_PY_METRICS_NAME_MAPPING[metric_name]
  except KeyError:
    raise ValueError('Invalid metric name %s' % metric_name)


def _get_stats_data(object measurement, object labels) -> _observability.StatsData:
  """Convert a Python measurement to StatsData.

  Args:
  measurement: A dict of type Mapping[str, Union[enum, Mapping[str, Union[float, int]]]]
    with keys and values as following:
      name -> cMetricsName
      type -> MeasurementType
      value -> {value_double: float | value_int: int}
  labels: Labels assciociated with stats data with type of dict[str, str].
  """
  measurement: Measurement
  labels: Mapping[str, str]

  metric_name = _cy_metric_name_to_py_metric_name(measurement['name'])
  if measurement['type'] == kMeasurementDouble:
    py_stat = _observability.StatsData(name=metric_name, measure_double=True,
                                       value_float=measurement['value']['value_double'],
                                       labels=labels)
  else:
    py_stat = _observability.StatsData(name=metric_name, measure_double=False,
                                       value_int=measurement['value']['value_int'],
                                       labels=labels)
  return py_stat


def _get_tracing_data(SpanCensusData span_data, vector[Label] span_labels,
                      vector[Annotation] span_annotations) -> _observability.TracingData:
  py_span_labels = _c_label_to_labels(span_labels)
  py_span_annotations = _c_annotation_to_annotations(span_annotations)
  return _observability.TracingData(name=_decode(span_data.name),
                                    start_time = _decode(span_data.start_time),
                                    end_time = _decode(span_data.end_time),
                                    trace_id = _decode(span_data.trace_id),
                                    span_id = _decode(span_data.span_id),
                                    parent_span_id = _decode(span_data.parent_span_id),
                                    status = _decode(span_data.status),
                                    should_sample = span_data.should_sample,
                                    child_span_count = span_data.child_span_count,
                                    span_labels = py_span_labels,
                                    span_annotations = py_span_annotations)


def _record_rpc_latency(object exporter, str method, float rpc_latency, str status_code) -> None:
  exporter: _observability.Exporter

  measurement = {}
  measurement['name'] = kRpcClientApiLatencyMeasureName
  measurement['type'] = kMeasurementDouble
  measurement['value'] = {'value_double': rpc_latency}

  labels = {}
  labels[_decode(kClientMethod)] = method.strip("/")
  labels[_decode(kClientStatus)] = status_code
  metric = _get_stats_data(measurement, labels)
  exporter.export_stats_data([metric])


cdef void _export_census_data(object exporter):
  """Main function running in export thread."""
  exporter: _observability.Exporter

  cdef int export_interval_ms = CENSUS_EXPORT_BATCH_INTERVAL_SECS * 1000
  while True:
    with nogil:
      while not GLOBAL_SHUTDOWN_EXPORT_THREAD:
        lk = new unique_lock[mutex](g_census_data_buffer_mutex)
        # Wait for next batch of census data OR timeout at fixed interval.
        # Batch export census data to minimize the time we acquiring the GIL.
        AwaitNextBatchLocked(dereference(lk), export_interval_ms)

        # Break only when buffer have data
        if not g_census_data_buffer.empty():
          del lk
          break
        else:
          del lk

    _flush_census_data(exporter)

    if GLOBAL_SHUTDOWN_EXPORT_THREAD:
      break # Break to shutdown exporting thead

  # Flush one last time before shutdown thread
  _flush_census_data(exporter)


cdef void _flush_census_data(object exporter):
  exporter: _observability.Exporter

  lk = new unique_lock[mutex](g_census_data_buffer_mutex)
  if g_census_data_buffer.empty():
    del lk
    return
  py_metrics_batch = []
  py_spans_batch = []
  while not g_census_data_buffer.empty():
    c_census_data = g_census_data_buffer.front()
    if c_census_data.type == kMetricData:
      py_labels = _c_label_to_labels(c_census_data.labels)
      py_measurement = _c_measurement_to_measurement(c_census_data.measurement_data)
      py_metric = _get_stats_data(py_measurement, py_labels)
      py_metrics_batch.append(py_metric)
    else:
      py_span = _get_tracing_data(c_census_data.span_data, c_census_data.span_data.span_labels,
                                  c_census_data.span_data.span_annotations)
      py_spans_batch.append(py_span)
    g_census_data_buffer.pop()

  del lk
  exporter.export_stats_data(py_metrics_batch)
  exporter.export_tracing_data(py_spans_batch)


cdef void _shutdown_exporting_thread():
  with nogil:
    global GLOBAL_SHUTDOWN_EXPORT_THREAD
    GLOBAL_SHUTDOWN_EXPORT_THREAD = True
    g_census_data_buffer_cv.notify_all()
  GLOBAL_EXPORT_THREAD.join(timeout=GRPC_PYTHON_CENSUS_EXPORT_THREAD_TIMEOUT)


cdef str _decode(bytes bytestring):
  if isinstance(bytestring, (str,)):
    return <str>bytestring
  else:
    try:
      return bytestring.decode('utf8')
    except UnicodeDecodeError:
      _LOGGER.exception('Invalid encoding on %s', bytestring)
      return bytestring.decode('latin1')



# Copyright 2020 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import datetime
import enum
import logging
from typing import Any, Dict, List, Optional, Set

from googleapiclient import discovery
import googleapiclient.errors

import framework.errors
from framework.helpers import retryers
from framework.infrastructure import gcp

logger = logging.getLogger(__name__)


class ComputeV1(
    gcp.api.GcpProjectApiResource
):  # pylint: disable=too-many-public-methods
    # TODO(sergiitk): move someplace better
    _WAIT_FOR_BACKEND_SEC = 60 * 10
    _WAIT_FOR_BACKEND_SLEEP_SEC = 4
    _WAIT_FOR_OPERATION_SEC = 60 * 10

    @dataclasses.dataclass(frozen=True)
    class GcpResource:
        name: str
        url: str

    @dataclasses.dataclass(frozen=True)
    class ZonalGcpResource(GcpResource):
        zone: str

    def __init__(
        self,
        api_manager: gcp.api.GcpApiManager,
        project: str,
        version: str = "v1",
    ):
        super().__init__(api_manager.compute(version), project)

    class HealthCheckProtocol(enum.Enum):
        TCP = enum.auto()
        GRPC = enum.auto()

    class BackendServiceProtocol(enum.Enum):
        HTTP2 = enum.auto()
        GRPC = enum.auto()

    def create_health_check(
        self,
        name: str,
        protocol: HealthCheckProtocol,
        *,
        port: Optional[int] = None,
    ) -> "GcpResource":
        if protocol is self.HealthCheckProtocol.TCP:
            health_check_field = "tcpHealthCheck"
        elif protocol is self.HealthCheckProtocol.GRPC:
            health_check_field = "grpcHealthCheck"
        else:
            raise TypeError(f"Unexpected Health Check protocol: {protocol}")

        health_check_settings = {}
        if port is None:
            health_check_settings["portSpecification"] = "USE_SERVING_PORT"
        else:
            health_check_settings["portSpecification"] = "USE_FIXED_PORT"
            health_check_settings["port"] = port

        return self._insert_resource(
            self.api.healthChecks(),
            {
                "name": name,
                "type": protocol.name,
                health_check_field: health_check_settings,
            },
        )

    def list_health_check(self):
        return self._list_resource(self.api.healthChecks())

    def delete_health_check(self, name):
        self._delete_resource(self.api.healthChecks(), "healthCheck", name)

    def create_firewall_rule(
        self,
        name: str,
        network_url: str,
        source_ranges: List[str],
        ports: List[str],
    ) -> Optional["GcpResource"]:
        try:
            return self._insert_resource(
                self.api.firewalls(),
                {
                    "allowed": [{"IPProtocol": "tcp", "ports": ports}],
                    "direction": "INGRESS",
                    "name": name,
                    "network": network_url,
                    "priority": 1000,
                    "sourceRanges": source_ranges,
                    "targetTags": ["allow-health-checks"],
                },
            )
        except googleapiclient.errors.HttpError as http_error:
            # TODO(lidiz) use status_code() when we upgrade googleapiclient
            if http_error.resp.status == 409:
                logger.debug("Firewall rule %s already existed", name)
                return None
            else:
                raise

    def delete_firewall_rule(self, name):
        self._delete_resource(self.api.firewalls(), "firewall", name)

    def create_backend_service_traffic_director(
        self,
        name: str,
        health_check: "GcpResource",
        affinity_header: Optional[str] = None,
        protocol: Optional[BackendServiceProtocol] = None,
        subset_size: Optional[int] = None,
        locality_lb_policies: Optional[List[dict]] = None,
        outlier_detection: Optional[dict] = None,
    ) -> "GcpResource":
        if not isinstance(protocol, self.BackendServiceProtocol):
            raise TypeError(f"Unexpected Backend Service protocol: {protocol}")
        body = {
            "name": name,
            "loadBalancingScheme": "INTERNAL_SELF_MANAGED",  # Traffic Director
            "healthChecks": [health_check.url],
            "protocol": protocol.name,
        }
        # If affinity header is specified, config the backend service to support
        # affinity, and set affinity header to the one given.
        if affinity_header:
            body["sessionAffinity"] = "HEADER_FIELD"
            body["localityLbPolicy"] = "RING_HASH"
            body["consistentHash"] = {
                "httpHeaderName": affinity_header,
            }
        if subset_size:
            body["subsetting"] = {
                "policy": "CONSISTENT_HASH_SUBSETTING",
                "subsetSize": subset_size,
            }
        if locality_lb_policies:
            body["localityLbPolicies"] = locality_lb_policies
        if outlier_detection:
            body["outlierDetection"] = outlier_detection
        return self._insert_resource(self.api.backendServices(), body)

    def get_backend_service_traffic_director(self, name: str) -> "GcpResource":
        return self._get_resource(
            self.api.backendServices(), backendService=name
        )

    def patch_backend_service(self, backend_service, body, **kwargs):
        self._patch_resource(
            collection=self.api.backendServices(),
            backendService=backend_service.name,
            body=body,
            **kwargs,
        )

    def backend_service_patch_backends(
        self,
        backend_service,
        backends,
        max_rate_per_endpoint: Optional[int] = None,
    ):
        if max_rate_per_endpoint is None:
            max_rate_per_endpoint = 5
        backend_list = [
            {
                "group": backend.url,
                "balancingMode": "RATE",
                "maxRatePerEndpoint": max_rate_per_endpoint,
            }
            for backend in backends
        ]

        self._patch_resource(
            collection=self.api.backendServices(),
            body={"backends": backend_list},
            backendService=backend_service.name,
        )

    def backend_service_remove_all_backends(self, backend_service):
        self._patch_resource(
            collection=self.api.backendServices(),
            body={"backends": []},
            backendService=backend_service.name,
        )

    def delete_backend_service(self, name):
        self._delete_resource(
            self.api.backendServices(), "backendService", name
        )

    def create_url_map(
        self,
        name: str,
        matcher_name: str,
        src_hosts,
        dst_default_backend_service: "GcpResource",
        dst_host_rule_match_backend_service: Optional["GcpResource"] = None,
    ) -> "GcpResource":
        if dst_host_rule_match_backend_service is None:
            dst_host_rule_match_backend_service = dst_default_backend_service
        return self._insert_resource(
            self.api.urlMaps(),
            {
                "name": name,
                "defaultService": dst_default_backend_service.url,
                "hostRules": [
                    {
                        "hosts": src_hosts,
                        "pathMatcher": matcher_name,
                    }
                ],
                "pathMatchers": [
                    {
                        "name": matcher_name,
                        "defaultService": dst_host_rule_match_backend_service.url,
                    }
                ],
            },
        )

    def create_url_map_with_content(self, url_map_body: Any) -> "GcpResource":
        return self._insert_resource(self.api.urlMaps(), url_map_body)

    def patch_url_map(self, url_map: "GcpResource", body, **kwargs):
        self._patch_resource(
            collection=self.api.urlMaps(),
            urlMap=url_map.name,
            body=body,
            **kwargs,
        )

    def delete_url_map(self, name):
        self._delete_resource(self.api.urlMaps(), "urlMap", name)

    def create_target_grpc_proxy(
        self,
        name: str,
        url_map: "GcpResource",
        validate_for_proxyless: bool = True,
    ) -> "GcpResource":
        return self._insert_resource(
            self.api.targetGrpcProxies(),
            {
                "name": name,
                "url_map": url_map.url,
                "validate_for_proxyless": validate_for_proxyless,
            },
        )

    def delete_target_grpc_proxy(self, name):
        self._delete_resource(
            self.api.targetGrpcProxies(), "targetGrpcProxy", name
        )

    def create_target_http_proxy(
        self,
        name: str,
        url_map: "GcpResource",
    ) -> "GcpResource":
        return self._insert_resource(
            self.api.targetHttpProxies(),
            {
                "name": name,
                "url_map": url_map.url,
            },
        )

    def delete_target_http_proxy(self, name):
        self._delete_resource(
            self.api.targetHttpProxies(), "targetHttpProxy", name
        )

    def create_forwarding_rule(
        self,
        name: str,
        src_port: int,
        target_proxy: "GcpResource",
        network_url: str,
        *,
        ip_address: str = "0.0.0.0",
    ) -> "GcpResource":
        return self._insert_resource(
            self.api.globalForwardingRules(),
            {
                "name": name,
                "loadBalancingScheme": "INTERNAL_SELF_MANAGED",  # Traffic Director
                "portRange": src_port,
                "IPAddress": ip_address,
                "network": network_url,
                "target": target_proxy.url,
            },
        )

    def exists_forwarding_rule(self, src_port) -> bool:
        # TODO(sergiitk): Better approach for confirming the port is available.
        #   It's possible a rule allocates actual port range, e.g 8000-9000,
        #   and this wouldn't catch it. For now, we assume there's no
        #   port ranges used in the project.
        filter_str = (
            f'(portRange eq "{src_port}-{src_port}") '
            '(IPAddress eq "0.0.0.0")'
            '(loadBalancingScheme eq "INTERNAL_SELF_MANAGED")'
        )
        return self._exists_resource(
            self.api.globalForwardingRules(), resource_filter=filter_str
        )

    def delete_forwarding_rule(self, name):
        self._delete_resource(
            self.api.globalForwardingRules(), "forwardingRule", name
        )

    def wait_for_network_endpoint_group(
        self,
        name: str,
        zone: str,
        *,
        timeout_sec=_WAIT_FOR_BACKEND_SEC,
        wait_sec=_WAIT_FOR_BACKEND_SLEEP_SEC,
    ):
        retryer = retryers.constant_retryer(
            wait_fixed=datetime.timedelta(seconds=wait_sec),
            timeout=datetime.timedelta(seconds=timeout_sec),
            check_result=lambda neg: neg and neg.get("size", 0) > 0,
        )
        network_endpoint_group = retryer(
            self._retry_load_network_endpoint_group, name, zone
        )
        # TODO(sergiitk): dataclass
        return self.ZonalGcpResource(
            network_endpoint_group["name"],
            network_endpoint_group["selfLink"],
            zone,
        )

    def _retry_load_network_endpoint_group(self, name: str, zone: str):
        try:
            neg = self.get_network_endpoint_group(name, zone)
            logger.debug(
                "Waiting for endpoints: NEG %s in zone %s, current count %s",
                neg["name"],
                zone,
                neg.get("size"),
            )
        except googleapiclient.errors.HttpError as error:
            # noinspection PyProtectedMember
            reason = error._get_reason()
            logger.debug(
                "Retrying NEG load, got %s, details %s",
                error.resp.status,
                reason,
            )
            raise
        return neg

    def get_network_endpoint_group(self, name, zone):
        neg = (
            self.api.networkEndpointGroups()
            .get(project=self.project, networkEndpointGroup=name, zone=zone)
            .execute()
        )
        # TODO(sergiitk): dataclass
        return neg

    def wait_for_backends_healthy_status(
        self,
        backend_service: GcpResource,
        backends: Set[ZonalGcpResource],
        *,
        timeout_sec: int = _WAIT_FOR_BACKEND_SEC,
        wait_sec: int = _WAIT_FOR_BACKEND_SLEEP_SEC,
    ) -> None:
        if not backends:
            raise ValueError("The list of backends to wait on is empty")

        timeout = datetime.timedelta(seconds=timeout_sec)
        retryer = retryers.constant_retryer(
            wait_fixed=datetime.timedelta(seconds=wait_sec),
            timeout=timeout,
            check_result=lambda result: result,
        )
        pending = set(backends)
        try:
            retryer(self._retry_backends_health, backend_service, pending)
        except retryers.RetryError as retry_err:
            unhealthy_backends: str = ",".join(
                [backend.name for backend in pending]
            )

            # Attempt to load backend health info for better debug info.
            try:
                unhealthy = []
                # Everything left in pending was unhealthy on the last retry.
                for backend in pending:
                    # It's possible the health status has changed since we
                    # gave up retrying, but this should be very rare.
                    health_status = self.get_backend_service_backend_health(
                        backend_service,
                        backend,
                    )
                    unhealthy.append(
                        {"name": backend.name, "health_status": health_status}
                    )

                # Override the plain list of unhealthy backend name with
                # the one showing the latest backend statuses.
                unhealthy_backends = self.resources_pretty_format(
                    unhealthy,
                    highlight=False,
                )
            except Exception as error:  # noqa pylint: disable=broad-except
                logger.debug(
                    "Couldn't load backend health info, plain list name"
                    "will be printed instead. Error: %r",
                    error,
                )

            retry_err.add_note(
                framework.errors.FrameworkError.note_blanket_error_info_below(
                    "One or several NEGs (Network Endpoint Groups) didn't"
                    " report HEALTHY status within expected timeout.",
                    info_below=(
                        f"Timeout {timeout} (h:mm:ss) waiting for backend"
                        f" service '{backend_service.name}' to report all NEGs"
                        " in the HEALTHY status:"
                        f" {[backend.name for backend in backends]}."
                        f"\nUnhealthy backends:\n{unhealthy_backends}"
                    ),
                )
            )

            raise

    def _retry_backends_health(
        self, backend_service: GcpResource, pending: Set[ZonalGcpResource]
    ):
        for backend in pending:
            result = self.get_backend_service_backend_health(
                backend_service, backend
            )
            if "healthStatus" not in result:
                logger.debug(
                    "Waiting for instances: backend %s, zone %s",
                    backend.name,
                    backend.zone,
                )
                continue

            backend_healthy = True
            for instance in result["healthStatus"]:
                logger.debug(
                    "Backend %s in zone %s: instance %s:%s health: %s",
                    backend.name,
                    backend.zone,
                    instance["ipAddress"],
                    instance["port"],
                    instance["healthState"],
                )
                if instance["healthState"] != "HEALTHY":
                    backend_healthy = False

            if backend_healthy:
                logger.info(
                    "Backend %s in zone %s reported healthy",
                    backend.name,
                    backend.zone,
                )
                pending.remove(backend)

        return not pending

    def get_backend_service_backend_health(self, backend_service, backend):
        return (
            self.api.backendServices()
            .getHealth(
                project=self.project,
                backendService=backend_service.name,
                body={"group": backend.url},
            )
            .execute()
        )

    def _get_resource(
        self, collection: discovery.Resource, **kwargs
    ) -> "GcpResource":
        resp = collection.get(project=self.project, **kwargs).execute()
        logger.info(
            "Loaded compute resource:\n%s", self.resource_pretty_format(resp)
        )
        return self.GcpResource(resp["name"], resp["selfLink"])

    def _exists_resource(
        self, collection: discovery.Resource, resource_filter: str
    ) -> bool:
        resp = collection.list(
            project=self.project, filter=resource_filter, maxResults=1
        ).execute(num_retries=self._GCP_API_RETRIES)
        if "kind" not in resp:
            # TODO(sergiitk): better error
            raise ValueError('List response "kind" is missing')
        return "items" in resp and resp["items"]

    def _insert_resource(
        self, collection: discovery.Resource, body: Dict[str, Any]
    ) -> "GcpResource":
        logger.info(
            "Creating compute resource:\n%s", self.resource_pretty_format(body)
        )
        resp = self._execute(collection.insert(project=self.project, body=body))
        return self.GcpResource(body["name"], resp["targetLink"])

    def _patch_resource(self, collection, body, **kwargs):
        logger.info(
            "Patching compute resource:\n%s", self.resource_pretty_format(body)
        )
        self._execute(
            collection.patch(project=self.project, body=body, **kwargs)
        )

    def _list_resource(self, collection: discovery.Resource):
        return collection.list(project=self.project).execute(
            num_retries=self._GCP_API_RETRIES
        )

    def _delete_resource(
        self,
        collection: discovery.Resource,
        resource_type: str,
        resource_name: str,
    ) -> bool:
        try:
            params = {"project": self.project, resource_type: resource_name}
            self._execute(collection.delete(**params))
            return True
        except googleapiclient.errors.HttpError as error:
            if error.resp and error.resp.status == 404:
                logger.info(
                    'Resource %s "%s" not deleted since it does not exist',
                    resource_type,
                    resource_name,
                )
            else:
                logger.warning(
                    'Failed to delete %s "%s", %r',
                    resource_type,
                    resource_name,
                    error,
                )
        return False

    @staticmethod
    def _operation_status_done(operation):
        return "status" in operation and operation["status"] == "DONE"

    def _execute(  # pylint: disable=arguments-differ
        self, request, *, timeout_sec=_WAIT_FOR_OPERATION_SEC
    ):
        operation = request.execute(num_retries=self._GCP_API_RETRIES)
        logger.debug("Operation %s", operation)
        return self._wait(operation["name"], timeout_sec)

    def _wait(
        self, operation_id: str, timeout_sec: int = _WAIT_FOR_OPERATION_SEC
    ) -> dict:
        logger.info(
            "Waiting %s sec for compute operation id: %s",
            timeout_sec,
            operation_id,
        )

        # TODO(sergiitk) try using wait() here
        # https://googleapis.github.io/google-api-python-client/docs/dyn/compute_v1.globalOperations.html#wait
        op_request = self.api.globalOperations().get(
            project=self.project, operation=operation_id
        )
        operation = self.wait_for_operation(
            operation_request=op_request,
            test_success_fn=self._operation_status_done,
            timeout_sec=timeout_sec,
        )

        logger.debug("Completed operation: %s", operation)
        if "error" in operation:
            # This shouldn't normally happen: gcp library raises on errors.
            raise Exception(
                f"Compute operation {operation_id} failed: {operation}"
            )
        return operation



# Copyright 2021 The gRPC Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A test framework built for urlMap related xDS test cases."""

import functools
import inspect
from typing import Any, Iterable, Mapping, Tuple

from absl import flags
from absl import logging

from framework import xds_flags
from framework import xds_k8s_flags
import framework.helpers.rand
from framework.infrastructure import gcp
from framework.infrastructure import k8s
from framework.infrastructure import traffic_director
from framework.test_app.runners.k8s import k8s_xds_client_runner
from framework.test_app.runners.k8s import k8s_xds_server_runner

flags.adopt_module_key_flags(xds_flags)
flags.adopt_module_key_flags(xds_k8s_flags)

STRATEGY = flags.DEFINE_enum(
    "strategy",
    default="reuse",
    enum_values=["create", "keep", "reuse"],
    help="Strategy of GCP resources management",
)

# Type alias
_KubernetesServerRunner = k8s_xds_server_runner.KubernetesServerRunner
_KubernetesClientRunner = k8s_xds_client_runner.KubernetesClientRunner
UrlMapType = Any
HostRule = Any
PathMatcher = Any


class _UrlMapChangeAggregator:
    """Where all the urlMap change happens."""

    def __init__(self, url_map_name: str):
        self._map = {
            "name": url_map_name,
            "defaultService": GcpResourceManager().default_backend_service(),
            "hostRules": [],
            "pathMatchers": [],
        }

    def get_map(self) -> UrlMapType:
        return self._map

    def apply_change(self, test_case: "XdsUrlMapTestCase") -> None:
        logging.info(
            "Apply urlMap change for test case: %s.%s",
            test_case.short_module_name,
            test_case.__name__,
        )
        url_map_parts = test_case.url_map_change(
            *self._get_test_case_url_map(test_case)
        )
        self._set_test_case_url_map(*url_map_parts)

    @staticmethod
    def _get_test_case_url_map(
        test_case: "XdsUrlMapTestCase",
    ) -> Tuple[HostRule, PathMatcher]:
        host_rule = {
            "hosts": [test_case.hostname()],
            "pathMatcher": test_case.path_matcher_name(),
        }
        path_matcher = {
            "name": test_case.path_matcher_name(),
            "defaultService": GcpResourceManager().default_backend_service(),
        }
        return host_rule, path_matcher

    def _set_test_case_url_map(
        self, host_rule: HostRule, path_matcher: PathMatcher
    ) -> None:
        self._map["hostRules"].append(host_rule)
        self._map["pathMatchers"].append(path_matcher)


def _package_flags() -> Mapping[str, Any]:
    """Automatically parse Abseil flags into a dictionary.

    Abseil flag is only available after the Abseil app initialization. If we use
    __new__ in our metaclass, the flag value parse will happen during the
    initialization of modules, hence will fail. That's why we are using __call__
    to inject metaclass magics, and the flag parsing will be delayed until the
    class is about to be instantiated.
    """
    res = {}
    for flag_module in [xds_flags, xds_k8s_flags]:
        for key, value in inspect.getmembers(flag_module):
            if isinstance(value, flags.FlagHolder):
                res[key.lower()] = value.value
    res["strategy"] = STRATEGY.value
    return res


class _MetaSingletonAndAbslFlags(type):
    """Ensures singleton and injects flag values."""

    # Allow different subclasses to create different singletons.
    _instances = {}
    # But we only parse Abseil flags once.
    _flags = None

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            if cls._flags is None:
                cls._flags = _package_flags()
            obj = super().__call__(cls._flags, *args, **kwargs)
            cls._instances[cls] = obj
            return obj
        return cls._instances[cls]


class GcpResourceManager(metaclass=_MetaSingletonAndAbslFlags):
    """Manages the lifecycle of GCP resources.

    The GCP resources including:
        - 3 K8s deployment (client, default backends, alternative backends)
        - Full set of the Traffic Director stuff
        - Merged gigantic urlMap from all imported test cases

    All resources are intended to be used across test cases and multiple runs
    (except the client K8s deployment).
    """

    # This class dynamically set, so disable "no-member" check.
    # pylint: disable=no-member

    def __init__(self, absl_flags: Mapping[str, Any] = None):
        if absl_flags is not None:
            for key in absl_flags:
                setattr(self, key, absl_flags[key])
        # Pick a client_namespace_suffix if not set
        if getattr(self, "resource_suffix", None) is None:
            self.resource_suffix = ""
        else:
            raise NotImplementedError(
                "Predefined resource_suffix is not supported for UrlMap tests"
            )
        logging.info(
            "GcpResourceManager: resource prefix=%s, suffix=%s",
            self.resource_prefix,
            self.resource_suffix,
        )

        # Must be called before KubernetesApiManager or GcpApiManager init.
        xds_flags.set_socket_default_timeout_from_flag()

        # API managers
        self.k8s_api_manager = k8s.KubernetesApiManager(self.kube_context)
        self.gcp_api_manager = gcp.api.GcpApiManager()
        self.td = traffic_director.TrafficDirectorManager(
            self.gcp_api_manager,
            self.project,
            resource_prefix=self.resource_prefix,
            resource_suffix=(self.resource_suffix or ""),
            network=self.network,
            compute_api_version=self.compute_api_version,
        )
        # Kubernetes namespace
        self.k8s_namespace = k8s.KubernetesNamespace(
            self.k8s_api_manager, self.resource_prefix
        )
        # Kubernetes Test Servers
        self.test_server_runner = _KubernetesServerRunner(
            self.k8s_namespace,
            deployment_name=self.server_name,
            image_name=self.server_image,
            gcp_project=self.project,
            gcp_api_manager=self.gcp_api_manager,
            gcp_service_account=self.gcp_service_account,
            td_bootstrap_image=self.td_bootstrap_image,
            xds_server_uri=self.xds_server_uri,
            network=self.network,
            enable_workload_identity=self.enable_workload_identity,
        )
        self.test_server_alternative_runner = _KubernetesServerRunner(
            self.k8s_namespace,
            deployment_name=self.server_name + "-alternative",
            image_name=self.server_image,
            gcp_project=self.project,
            gcp_api_manager=self.gcp_api_manager,
            gcp_service_account=self.gcp_service_account,
            td_bootstrap_image=self.td_bootstrap_image,
            xds_server_uri=self.xds_server_uri,
            network=self.network,
            enable_workload_identity=self.enable_workload_identity,
            reuse_namespace=True,
        )
        self.test_server_affinity_runner = _KubernetesServerRunner(
            self.k8s_namespace,
            deployment_name=self.server_name + "-affinity",
            image_name=self.server_image,
            gcp_project=self.project,
            gcp_api_manager=self.gcp_api_manager,
            gcp_service_account=self.gcp_service_account,
            td_bootstrap_image=self.td_bootstrap_image,
            xds_server_uri=self.xds_server_uri,
            network=self.network,
            enable_workload_identity=self.enable_workload_identity,
            reuse_namespace=True,
        )
        logging.info("Strategy of GCP resources management: %s", self.strategy)

    def create_test_client_runner(self):
        if self.resource_suffix:
            client_namespace_suffix = self.resource_suffix
        else:
            client_namespace_suffix = (
                framework.helpers.rand.random_resource_suffix()
            )
        logging.info(
            "GcpResourceManager: client_namespace_suffix=%s",
            client_namespace_suffix,
        )
        # Kubernetes Test Client
        namespace_name = _KubernetesClientRunner.make_namespace_name(
            self.resource_prefix, client_namespace_suffix
        )
        return _KubernetesClientRunner(
            k8s.KubernetesNamespace(self.k8s_api_manager, namespace_name),
            deployment_name=self.client_name,
            image_name=self.client_image,
            gcp_project=self.project,
            gcp_api_manager=self.gcp_api_manager,
            gcp_service_account=self.gcp_service_account,
            td_bootstrap_image=self.td_bootstrap_image,
            xds_server_uri=self.xds_server_uri,
            network=self.network,
            debug_use_port_forwarding=self.debug_use_port_forwarding,
            enable_workload_identity=self.enable_workload_identity,
            stats_port=self.client_port,
        )

    def _pre_cleanup(self):
        # Cleanup existing debris
        logging.info("GcpResourceManager: pre clean-up")
        self.td.cleanup(force=True)
        self.test_server_runner.delete_namespace()

    def setup(self, test_case_classes: Iterable["XdsUrlMapTestCase"]) -> None:
        if self.strategy not in ["create", "keep"]:
            logging.info(
                "GcpResourceManager: skipping setup for strategy [%s]",
                self.strategy,
            )
            return
        # Clean up debris from previous runs
        self._pre_cleanup()
        # Start creating GCP resources
        logging.info("GcpResourceManager: start setup")
        # Firewall
        if self.ensure_firewall:
            self.td.create_firewall_rule(
                allowed_ports=self.firewall_allowed_ports
            )
        # Health Checks
        self.td.create_health_check()
        # Backend Services
        self.td.create_backend_service()
        self.td.create_alternative_backend_service()
        self.td.create_affinity_backend_service()
        # Construct UrlMap from test classes
        aggregator = _UrlMapChangeAggregator(
            url_map_name=self.td.make_resource_name(self.td.URL_MAP_NAME)
        )
        for test_case_class in test_case_classes:
            aggregator.apply_change(test_case_class)
        final_url_map = aggregator.get_map()
        # UrlMap
        self.td.create_url_map_with_content(final_url_map)
        # Target Proxy
        self.td.create_target_proxy()
        # Forwarding Rule
        self.td.create_forwarding_rule(self.server_xds_port)
        # Kubernetes Test Server
        self.test_server_runner.run(
            test_port=self.server_port,
            maintenance_port=self.server_maintenance_port,
        )
        # Kubernetes Test Server Alternative
        self.test_server_alternative_runner.run(
            test_port=self.server_port,
            maintenance_port=self.server_maintenance_port,
        )
        # Kubernetes Test Server Affinity. 3 endpoints to test that only the
        # picked sub-channel is connected.
        self.test_server_affinity_runner.run(
            test_port=self.server_port,
            maintenance_port=self.server_maintenance_port,
            replica_count=3,
        )
        # Add backend to default backend service
        neg_name, neg_zones = self.k8s_namespace.get_service_neg(
            self.test_server_runner.service_name, self.server_port
        )
        self.td.backend_service_add_neg_backends(neg_name, neg_zones)
        # Add backend to alternative backend service
        neg_name_alt, neg_zones_alt = self.k8s_namespace.get_service_neg(
            self.test_server_alternative_runner.service_name, self.server_port
        )
        self.td.alternative_backend_service_add_neg_backends(
            neg_name_alt, neg_zones_alt
        )
        # Add backend to affinity backend service
        (
            neg_name_affinity,
            neg_zones_affinity,
        ) = self.k8s_namespace.get_service_neg(
            self.test_server_affinity_runner.service_name, self.server_port
        )
        self.td.affinity_backend_service_add_neg_backends(
            neg_name_affinity, neg_zones_affinity
        )
        # Wait for healthy backends
        self.td.wait_for_backends_healthy_status()
        self.td.wait_for_alternative_backends_healthy_status()
        self.td.wait_for_affinity_backends_healthy_status()

    def cleanup(self) -> None:
        if self.strategy not in ["create"]:
            logging.info(
                "GcpResourceManager: skipping tear down for strategy [%s]",
                self.strategy,
            )
            return
        logging.info("GcpResourceManager: start tear down")
        if hasattr(self, "td"):
            self.td.cleanup(force=True)
        if hasattr(self, "test_server_runner"):
            self.test_server_runner.cleanup(force=True)
        if hasattr(self, "test_server_alternative_runner"):
            self.test_server_alternative_runner.cleanup(
                force=True, force_namespace=True
            )
        if hasattr(self, "test_server_affinity_runner"):
            self.test_server_affinity_runner.cleanup(
                force=True, force_namespace=True
            )

    @functools.lru_cache(None)
    def default_backend_service(self) -> str:
        """Returns default backend service URL."""
        self.td.load_backend_service()
        return self.td.backend_service.url

    @functools.lru_cache(None)
    def alternative_backend_service(self) -> str:
        """Returns alternative backend service URL."""
        self.td.load_alternative_backend_service()
        return self.td.alternative_backend_service.url

    @functools.lru_cache(None)
    def affinity_backend_service(self) -> str:
        """Returns affinity backend service URL."""
        self.td.load_affinity_backend_service()
        return self.td.affinity_backend_service.url



# Copyright 2020 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import dataclasses
import logging
from typing import Any, Dict

from google.rpc import code_pb2
import tenacity

from framework.infrastructure import gcp

logger = logging.getLogger(__name__)

# Type aliases
GcpResource = gcp.compute.ComputeV1.GcpResource


@dataclasses.dataclass(frozen=True)
class ServerTlsPolicy:
    url: str
    name: str
    server_certificate: dict
    mtls_policy: dict
    update_time: str
    create_time: str

    @classmethod
    def from_response(
        cls, name: str, response: Dict[str, Any]
    ) -> "ServerTlsPolicy":
        return cls(
            name=name,
            url=response["name"],
            server_certificate=response.get("serverCertificate", {}),
            mtls_policy=response.get("mtlsPolicy", {}),
            create_time=response["createTime"],
            update_time=response["updateTime"],
        )


@dataclasses.dataclass(frozen=True)
class ClientTlsPolicy:
    url: str
    name: str
    client_certificate: dict
    server_validation_ca: list
    update_time: str
    create_time: str

    @classmethod
    def from_response(
        cls, name: str, response: Dict[str, Any]
    ) -> "ClientTlsPolicy":
        return cls(
            name=name,
            url=response["name"],
            client_certificate=response.get("clientCertificate", {}),
            server_validation_ca=response.get("serverValidationCa", []),
            create_time=response["createTime"],
            update_time=response["updateTime"],
        )


@dataclasses.dataclass(frozen=True)
class AuthorizationPolicy:
    url: str
    name: str
    update_time: str
    create_time: str
    action: str
    rules: list

    @classmethod
    def from_response(
        cls, name: str, response: Dict[str, Any]
    ) -> "AuthorizationPolicy":
        return cls(
            name=name,
            url=response["name"],
            create_time=response["createTime"],
            update_time=response["updateTime"],
            action=response["action"],
            rules=response.get("rules", []),
        )


class _NetworkSecurityBase(
    gcp.api.GcpStandardCloudApiResource, metaclass=abc.ABCMeta
):
    """Base class for NetworkSecurity APIs."""

    # TODO(https://github.com/grpc/grpc/issues/29532) remove pylint disable
    # pylint: disable=abstract-method

    def __init__(self, api_manager: gcp.api.GcpApiManager, project: str):
        super().__init__(api_manager.networksecurity(self.api_version), project)
        # Shortcut to projects/*/locations/ endpoints
        self._api_locations = self.api.projects().locations()

    @property
    def api_name(self) -> str:
        return "networksecurity"

    def _execute(
        self, *args, **kwargs
    ):  # pylint: disable=signature-differs,arguments-differ
        # Workaround TD bug: throttled operations are reported as internal.
        # Ref b/175345578
        retryer = tenacity.Retrying(
            retry=tenacity.retry_if_exception(self._operation_internal_error),
            wait=tenacity.wait_fixed(10),
            stop=tenacity.stop_after_delay(5 * 60),
            before_sleep=tenacity.before_sleep_log(logger, logging.DEBUG),
            reraise=True,
        )
        retryer(super()._execute, *args, **kwargs)

    @staticmethod
    def _operation_internal_error(exception):
        return (
            isinstance(exception, gcp.api.OperationError)
            and exception.error.code == code_pb2.INTERNAL
        )


class NetworkSecurityV1Beta1(_NetworkSecurityBase):
    """NetworkSecurity API v1beta1."""

    SERVER_TLS_POLICIES = "serverTlsPolicies"
    CLIENT_TLS_POLICIES = "clientTlsPolicies"
    AUTHZ_POLICIES = "authorizationPolicies"

    @property
    def api_version(self) -> str:
        return "v1beta1"

    def create_server_tls_policy(self, name: str, body: dict) -> GcpResource:
        return self._create_resource(
            collection=self._api_locations.serverTlsPolicies(),
            body=body,
            serverTlsPolicyId=name,
        )

    def get_server_tls_policy(self, name: str) -> ServerTlsPolicy:
        response = self._get_resource(
            collection=self._api_locations.serverTlsPolicies(),
            full_name=self.resource_full_name(name, self.SERVER_TLS_POLICIES),
        )
        return ServerTlsPolicy.from_response(name, response)

    def delete_server_tls_policy(self, name: str) -> bool:
        return self._delete_resource(
            collection=self._api_locations.serverTlsPolicies(),
            full_name=self.resource_full_name(name, self.SERVER_TLS_POLICIES),
        )

    def create_client_tls_policy(self, name: str, body: dict) -> GcpResource:
        return self._create_resource(
            collection=self._api_locations.clientTlsPolicies(),
            body=body,
            clientTlsPolicyId=name,
        )

    def get_client_tls_policy(self, name: str) -> ClientTlsPolicy:
        response = self._get_resource(
            collection=self._api_locations.clientTlsPolicies(),
            full_name=self.resource_full_name(name, self.CLIENT_TLS_POLICIES),
        )
        return ClientTlsPolicy.from_response(name, response)

    def delete_client_tls_policy(self, name: str) -> bool:
        return self._delete_resource(
            collection=self._api_locations.clientTlsPolicies(),
            full_name=self.resource_full_name(name, self.CLIENT_TLS_POLICIES),
        )

    def create_authz_policy(self, name: str, body: dict) -> GcpResource:
        return self._create_resource(
            collection=self._api_locations.authorizationPolicies(),
            body=body,
            authorizationPolicyId=name,
        )

    def get_authz_policy(self, name: str) -> ClientTlsPolicy:
        response = self._get_resource(
            collection=self._api_locations.authorizationPolicies(),
            full_name=self.resource_full_name(name, self.AUTHZ_POLICIES),
        )
        return ClientTlsPolicy.from_response(name, response)

    def delete_authz_policy(self, name: str) -> bool:
        return self._delete_resource(
            collection=self._api_locations.authorizationPolicies(),
            full_name=self.resource_full_name(name, self.AUTHZ_POLICIES),
        )


class NetworkSecurityV1Alpha1(NetworkSecurityV1Beta1):
    """NetworkSecurity API v1alpha1.

    Note: extending v1beta1 class presumes that v1beta1 is just a v1alpha1 API
    graduated into a more stable version. This is true in most cases. However,
    v1alpha1 class can always override and reimplement incompatible methods.
    """

    @property
    def api_version(self) -> str:
        return "v1alpha1"



# Copyright 2023 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from concurrent import futures
import json
import logging
import os
import random
from typing import Any, Dict, List
import unittest

import grpc
import grpc_observability
from grpc_observability import _cyobservability
from grpc_observability import _observability

logger = logging.getLogger(__name__)

_REQUEST = b"\x00\x00\x00"
_RESPONSE = b"\x00\x00\x00"

_UNARY_UNARY = "/test/UnaryUnary"
_UNARY_STREAM = "/test/UnaryStream"
_STREAM_UNARY = "/test/StreamUnary"
_STREAM_STREAM = "/test/StreamStream"
STREAM_LENGTH = 5
TRIGGER_RPC_METADATA = ("control", "trigger_rpc")
TRIGGER_RPC_TO_NEW_SERVER_METADATA = ("to_new_server", "")

CONFIG_ENV_VAR_NAME = "GRPC_GCP_OBSERVABILITY_CONFIG"
CONFIG_FILE_ENV_VAR_NAME = "GRPC_GCP_OBSERVABILITY_CONFIG_FILE"

_VALID_CONFIG_TRACING_STATS = {
    "project_id": "test-project",
    "cloud_trace": {"sampling_rate": 1.00},
    "cloud_monitoring": {},
}
_VALID_CONFIG_TRACING_ONLY = {
    "project_id": "test-project",
    "cloud_trace": {"sampling_rate": 1.00},
}
_VALID_CONFIG_STATS_ONLY = {
    "project_id": "test-project",
    "cloud_monitoring": {},
}
_VALID_CONFIG_STATS_ONLY_STR = """
{
    'project_id': 'test-project',
    'cloud_monitoring': {}
}
"""
# Depends on grpc_core::IsTransportSuppliesClientLatencyEnabled,
# the following metrcis might not exist.
_SKIP_VEFIRY = [_cyobservability.MetricsName.CLIENT_TRANSPORT_LATENCY]
_SPAN_PREFIXS = ["Recv", "Sent", "Attempt"]


class TestExporter(_observability.Exporter):
    def __init__(
        self,
        metrics: List[_observability.StatsData],
        spans: List[_observability.TracingData],
    ):
        self.span_collecter = spans
        self.metric_collecter = metrics
        self._server = None

    def export_stats_data(
        self, stats_data: List[_observability.StatsData]
    ) -> None:
        self.metric_collecter.extend(stats_data)

    def export_tracing_data(
        self, tracing_data: List[_observability.TracingData]
    ) -> None:
        self.span_collecter.extend(tracing_data)


def handle_unary_unary(request, servicer_context):
    if TRIGGER_RPC_METADATA in servicer_context.invocation_metadata():
        for k, v in servicer_context.invocation_metadata():
            if "port" in k:
                unary_unary_call(port=int(v))
            if "to_new_server" in k:
                second_server = grpc.server(
                    futures.ThreadPoolExecutor(max_workers=10)
                )
                second_server.add_generic_rpc_handlers((_GenericHandler(),))
                second_server_port = second_server.add_insecure_port("[::]:0")
                second_server.start()
                unary_unary_call(port=second_server_port)
                second_server.stop(0)
    return _RESPONSE


def handle_unary_stream(request, servicer_context):
    for _ in range(STREAM_LENGTH):
        yield _RESPONSE


def handle_stream_unary(request_iterator, servicer_context):
    return _RESPONSE


def handle_stream_stream(request_iterator, servicer_context):
    for request in request_iterator:
        yield _RESPONSE


class _MethodHandler(grpc.RpcMethodHandler):
    def __init__(self, request_streaming, response_streaming):
        self.request_streaming = request_streaming
        self.response_streaming = response_streaming
        self.request_deserializer = None
        self.response_serializer = None
        self.unary_unary = None
        self.unary_stream = None
        self.stream_unary = None
        self.stream_stream = None
        if self.request_streaming and self.response_streaming:
            self.stream_stream = lambda x, y: handle_stream_stream(x, y)
        elif self.request_streaming:
            self.stream_unary = lambda x, y: handle_stream_unary(x, y)
        elif self.response_streaming:
            self.unary_stream = lambda x, y: handle_unary_stream(x, y)
        else:
            self.unary_unary = lambda x, y: handle_unary_unary(x, y)


class _GenericHandler(grpc.GenericRpcHandler):
    def service(self, handler_call_details):
        if handler_call_details.method == _UNARY_UNARY:
            return _MethodHandler(False, False)
        elif handler_call_details.method == _UNARY_STREAM:
            return _MethodHandler(False, True)
        elif handler_call_details.method == _STREAM_UNARY:
            return _MethodHandler(True, False)
        elif handler_call_details.method == _STREAM_STREAM:
            return _MethodHandler(True, True)
        else:
            return None


class ObservabilityTest(unittest.TestCase):
    def setUp(self):
        self.all_metric = []
        self.all_span = []
        self.test_exporter = TestExporter(self.all_metric, self.all_span)
        self._server = None
        self._port = None

    def tearDown(self):
        os.environ[CONFIG_ENV_VAR_NAME] = ""
        os.environ[CONFIG_FILE_ENV_VAR_NAME] = ""
        if self._server:
            self._server.stop(0)

    def testRecordUnaryUnary(self):
        self._set_config_file(_VALID_CONFIG_TRACING_STATS)
        with grpc_observability.GCPOpenCensusObservability(
            exporter=self.test_exporter
        ):
            self._start_server()
            unary_unary_call(port=self._port)

        self.assertGreater(len(self.all_metric), 0)
        self.assertGreater(len(self.all_span), 0)
        self._validate_metrics(self.all_metric)
        self._validate_spans(self.all_span)

    def testContextPropagationToSameServer(self):
        # Sends two RPCs, one from gRPC client and the other from gRPC server:
        #   gRPC Client -> gRPC Server 1 -> gRPC Server 1
        # Verify that the trace_id was propagated to the 2nd RPC.
        self._set_config_file(_VALID_CONFIG_TRACING_ONLY)
        with grpc_observability.GCPOpenCensusObservability(
            exporter=self.test_exporter
        ):
            port = self._start_server()
            metadata = (
                TRIGGER_RPC_METADATA,
                ("port", str(port)),
            )
            unary_unary_call(port=port, metadata=metadata)

        # 2 of each for ["Recv", "Sent", "Attempt"]
        self.assertEqual(len(self.all_span), 6)
        trace_id = self.all_span[0].trace_id
        for span in self.all_span:
            self.assertEqual(span.trace_id, trace_id)

    def testContextPropagationToNewServer(self):
        # Sends two RPCs, one from gRPC client and the other from gRPC server:
        #   gRPC Client -> gRPC Server 1 -> gRPC Server 2
        # Verify that the trace_id was propagated to the 2nd RPC.
        # This test case is to make sure that the context from one thread can
        # be propagated to different thread.
        self._set_config_file(_VALID_CONFIG_TRACING_ONLY)
        with grpc_observability.GCPOpenCensusObservability(
            exporter=self.test_exporter
        ):
            port = self._start_server()
            metadata = (
                TRIGGER_RPC_METADATA,
                TRIGGER_RPC_TO_NEW_SERVER_METADATA,
            )
            unary_unary_call(port=port, metadata=metadata)

        # 2 of each for ["Recv", "Sent", "Attempt"]
        self.assertEqual(len(self.all_span), 6)
        trace_id = self.all_span[0].trace_id
        for span in self.all_span:
            self.assertEqual(span.trace_id, trace_id)

    def testThrowErrorWithoutConfig(self):
        with self.assertRaises(ValueError):
            with grpc_observability.GCPOpenCensusObservability(
                exporter=self.test_exporter
            ):
                pass

    def testThrowErrorWithInvalidConfig(self):
        _INVALID_CONFIG = "INVALID"
        self._set_config_file(_INVALID_CONFIG)
        with self.assertRaises(ValueError):
            with grpc_observability.GCPOpenCensusObservability(
                exporter=self.test_exporter
            ):
                pass

    def testNoErrorAndDataWithEmptyConfig(self):
        _EMPTY_CONFIG = {}
        self._set_config_file(_EMPTY_CONFIG)
        # Empty config still require project_id
        os.environ["GCP_PROJECT"] = "test-project"
        with grpc_observability.GCPOpenCensusObservability(
            exporter=self.test_exporter
        ):
            self._start_server()
            unary_unary_call(port=self._port)

        self.assertEqual(len(self.all_metric), 0)
        self.assertEqual(len(self.all_span), 0)

    def testThrowErrorWhenCallingMultipleInit(self):
        self._set_config_file(_VALID_CONFIG_TRACING_STATS)
        with self.assertRaises(ValueError):
            with grpc_observability.GCPOpenCensusObservability(
                exporter=self.test_exporter
            ) as o11y:
                grpc._observability.observability_init(o11y)

    def testRecordUnaryUnaryStatsOnly(self):
        self._set_config_file(_VALID_CONFIG_STATS_ONLY)
        with grpc_observability.GCPOpenCensusObservability(
            exporter=self.test_exporter
        ):
            self._start_server()
            unary_unary_call(port=self._port)

        self.assertEqual(len(self.all_span), 0)
        self.assertGreater(len(self.all_metric), 0)
        self._validate_metrics(self.all_metric)

    def testRecordUnaryUnaryTracingOnly(self):
        self._set_config_file(_VALID_CONFIG_TRACING_ONLY)
        with grpc_observability.GCPOpenCensusObservability(
            exporter=self.test_exporter
        ):
            self._start_server()
            unary_unary_call(port=self._port)

        self.assertEqual(len(self.all_metric), 0)
        self.assertGreater(len(self.all_span), 0)
        self._validate_spans(self.all_span)

    def testRecordUnaryStream(self):
        self._set_config_file(_VALID_CONFIG_TRACING_STATS)
        with grpc_observability.GCPOpenCensusObservability(
            exporter=self.test_exporter
        ):
            self._start_server()
            unary_stream_call(port=self._port)

        self.assertGreater(len(self.all_metric), 0)
        self.assertGreater(len(self.all_span), 0)
        self._validate_metrics(self.all_metric)
        self._validate_spans(self.all_span)

    def testRecordStreamUnary(self):
        self._set_config_file(_VALID_CONFIG_TRACING_STATS)
        with grpc_observability.GCPOpenCensusObservability(
            exporter=self.test_exporter
        ):
            self._start_server()
            stream_unary_call(port=self._port)

        self.assertTrue(len(self.all_metric) > 0)
        self.assertTrue(len(self.all_span) > 0)
        self._validate_metrics(self.all_metric)
        self._validate_spans(self.all_span)

    def testRecordStreamStream(self):
        self._set_config_file(_VALID_CONFIG_TRACING_STATS)
        with grpc_observability.GCPOpenCensusObservability(
            exporter=self.test_exporter
        ):
            self._start_server()
            stream_stream_call(port=self._port)

        self.assertGreater(len(self.all_metric), 0)
        self.assertGreater(len(self.all_span), 0)
        self._validate_metrics(self.all_metric)
        self._validate_spans(self.all_span)

    def testNoRecordBeforeInit(self):
        self._set_config_file(_VALID_CONFIG_TRACING_STATS)
        self._start_server()
        unary_unary_call(port=self._port)
        self.assertEqual(len(self.all_metric), 0)
        self.assertEqual(len(self.all_span), 0)
        self._server.stop(0)

        with grpc_observability.GCPOpenCensusObservability(
            exporter=self.test_exporter
        ):
            self._start_server()
            unary_unary_call(port=self._port)

        self.assertGreater(len(self.all_metric), 0)
        self.assertGreater(len(self.all_span), 0)
        self._validate_metrics(self.all_metric)
        self._validate_spans(self.all_span)

    def testNoRecordAfterExit(self):
        self._set_config_file(_VALID_CONFIG_TRACING_STATS)
        with grpc_observability.GCPOpenCensusObservability(
            exporter=self.test_exporter
        ):
            self._start_server()
            unary_unary_call(port=self._port)

        self.assertGreater(len(self.all_metric), 0)
        self.assertGreater(len(self.all_span), 0)
        current_metric_len = len(self.all_metric)
        current_spans_len = len(self.all_span)
        self._validate_metrics(self.all_metric)
        self._validate_spans(self.all_span)

        unary_unary_call(port=self._port)
        self.assertEqual(len(self.all_metric), current_metric_len)
        self.assertEqual(len(self.all_span), current_spans_len)

    def testTraceSamplingRate(self):
        # Make 40 UnaryCall's
        # With 50% sampling rate, we should get 10-30 traces with >99.93% probability
        # Each trace will have three span (Send, Recv, Attempt)
        _CALLS = 40
        _LOWER_BOUND = 10 * 3
        _HIGHER_BOUND = 30 * 3
        _VALID_CONFIG_TRACING_ONLY_SAMPLE_HALF = {
            "project_id": "test-project",
            "cloud_trace": {"sampling_rate": 0.5},
        }
        self._set_config_file(_VALID_CONFIG_TRACING_ONLY_SAMPLE_HALF)
        with grpc_observability.GCPOpenCensusObservability(
            exporter=self.test_exporter
        ):
            self._start_server()
            for _ in range(_CALLS):
                unary_unary_call(port=self._port)
        self.assertEqual(len(self.all_metric), 0)
        self.assertGreaterEqual(len(self.all_span), _LOWER_BOUND)
        self.assertLessEqual(len(self.all_span), _HIGHER_BOUND)
        self._validate_spans(self.all_span)

    def testConfigFileOverEnvVar(self):
        # env var have only stats enabled
        os.environ[CONFIG_ENV_VAR_NAME] = _VALID_CONFIG_STATS_ONLY_STR
        # config_file have only tracing enabled
        self._set_config_file(_VALID_CONFIG_TRACING_ONLY)

        with grpc_observability.GCPOpenCensusObservability(
            exporter=self.test_exporter
        ):
            self._start_server()
            unary_unary_call(port=self._port)

        self.assertEqual(len(self.all_metric), 0)
        self.assertGreater(len(self.all_span), 0)
        self._validate_spans(self.all_span)

    def _set_config_file(self, config: Dict[str, Any]) -> None:
        # Using random name here so multiple tests can run with different config files.
        config_file_path = "/tmp/" + str(random.randint(0, 100000))
        with open(config_file_path, "w", encoding="utf-8") as f:
            f.write(json.dumps(config))
        os.environ[CONFIG_FILE_ENV_VAR_NAME] = config_file_path

    def _start_server(self) -> None:
        self._server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
        self._server.add_generic_rpc_handlers((_GenericHandler(),))
        self._port = self._server.add_insecure_port("[::]:0")
        self._server.start()
        return self._port

    def _validate_metrics(
        self, metrics: List[_observability.StatsData]
    ) -> None:
        metric_names = set(metric.name for metric in metrics)
        for name in _cyobservability.MetricsName:
            if name in _SKIP_VEFIRY:
                continue
            if name not in metric_names:
                logger.error(
                    "metric %s not found in exported metrics: %s!",
                    name,
                    metric_names,
                )
            self.assertTrue(name in metric_names)

    def _validate_spans(
        self, tracing_data: List[_observability.TracingData]
    ) -> None:
        span_names = set(data.name for data in tracing_data)
        for prefix in _SPAN_PREFIXS:
            prefix_exist = any(prefix in name for name in span_names)
            if not prefix_exist:
                logger.error(
                    "missing span with prefix %s in exported spans: %s!",
                    prefix,
                    span_names,
                )
            self.assertTrue(prefix_exist)


def unary_unary_call(port, metadata=None):
    with grpc.insecure_channel(f"localhost:{port}") as channel:
        multi_callable = channel.unary_unary(_UNARY_UNARY)
        if metadata:
            unused_response, call = multi_callable.with_call(
                _REQUEST, metadata=metadata
            )
        else:
            unused_response, call = multi_callable.with_call(_REQUEST)


def unary_stream_call(port):
    with grpc.insecure_channel(f"localhost:{port}") as channel:
        multi_callable = channel.unary_stream(_UNARY_STREAM)
        call = multi_callable(_REQUEST)
        for _ in call:
            pass


def stream_unary_call(port):
    with grpc.insecure_channel(f"localhost:{port}") as channel:
        multi_callable = channel.stream_unary(_STREAM_UNARY)
        unused_response, call = multi_callable.with_call(
            iter([_REQUEST] * STREAM_LENGTH)
        )


def stream_stream_call(port):
    with grpc.insecure_channel(f"localhost:{port}") as channel:
        multi_callable = channel.stream_stream(_STREAM_STREAM)
        call = multi_callable(iter([_REQUEST] * STREAM_LENGTH))
        for _ in call:
            pass


if __name__ == "__main__":
    logging.basicConfig()
    unittest.main(verbosity=2)



# Copyright 2016 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import unittest

import grpc

from tests.unit import test_common
from tests.unit.framework.common import test_constants
from tests.unit.framework.common import test_control

_SERIALIZE_REQUEST = lambda bytestring: bytestring * 2
_DESERIALIZE_REQUEST = lambda bytestring: bytestring[len(bytestring) // 2 :]
_SERIALIZE_RESPONSE = lambda bytestring: bytestring * 3
_DESERIALIZE_RESPONSE = lambda bytestring: bytestring[: len(bytestring) // 3]

_UNARY_UNARY = "/test/UnaryUnary"
_UNARY_UNARY_NESTED_EXCEPTION = "/test/UnaryUnaryNestedException"
_UNARY_STREAM = "/test/UnaryStream"
_STREAM_UNARY = "/test/StreamUnary"
_STREAM_STREAM = "/test/StreamStream"
_DEFECTIVE_GENERIC_RPC_HANDLER = "/test/DefectiveGenericRpcHandler"


class _Handler(object):
    def __init__(self, control):
        self._control = control

    def handle_unary_unary(self, request, servicer_context):
        self._control.control()
        if servicer_context is not None:
            servicer_context.set_trailing_metadata(
                (
                    (
                        "testkey",
                        "testvalue",
                    ),
                )
            )
        return request

    def handle_unary_unary_with_nested_exception(
        self, request, servicer_context
    ):
        raise test_control.NestedDefect()

    def handle_unary_stream(self, request, servicer_context):
        for _ in range(test_constants.STREAM_LENGTH):
            self._control.control()
            yield request
        self._control.control()
        if servicer_context is not None:
            servicer_context.set_trailing_metadata(
                (
                    (
                        "testkey",
                        "testvalue",
                    ),
                )
            )

    def handle_stream_unary(self, request_iterator, servicer_context):
        if servicer_context is not None:
            servicer_context.invocation_metadata()
        self._control.control()
        response_elements = []
        for request in request_iterator:
            self._control.control()
            response_elements.append(request)
        self._control.control()
        if servicer_context is not None:
            servicer_context.set_trailing_metadata(
                (
                    (
                        "testkey",
                        "testvalue",
                    ),
                )
            )
        return b"".join(response_elements)

    def handle_stream_stream(self, request_iterator, servicer_context):
        self._control.control()
        if servicer_context is not None:
            servicer_context.set_trailing_metadata(
                (
                    (
                        "testkey",
                        "testvalue",
                    ),
                )
            )
        for request in request_iterator:
            self._control.control()
            yield request
        self._control.control()

    def defective_generic_rpc_handler(self):
        raise test_control.Defect()


class _MethodHandler(grpc.RpcMethodHandler):
    def __init__(
        self,
        request_streaming,
        response_streaming,
        request_deserializer,
        response_serializer,
        unary_unary,
        unary_stream,
        stream_unary,
        stream_stream,
    ):
        self.request_streaming = request_streaming
        self.response_streaming = response_streaming
        self.request_deserializer = request_deserializer
        self.response_serializer = response_serializer
        self.unary_unary = unary_unary
        self.unary_stream = unary_stream
        self.stream_unary = stream_unary
        self.stream_stream = stream_stream


class _GenericHandler(grpc.GenericRpcHandler):
    def __init__(self, handler):
        self._handler = handler

    def service(self, handler_call_details):
        if handler_call_details.method == _UNARY_UNARY:
            return _MethodHandler(
                False,
                False,
                None,
                None,
                self._handler.handle_unary_unary,
                None,
                None,
                None,
            )
        elif handler_call_details.method == _UNARY_STREAM:
            return _MethodHandler(
                False,
                True,
                _DESERIALIZE_REQUEST,
                _SERIALIZE_RESPONSE,
                None,
                self._handler.handle_unary_stream,
                None,
                None,
            )
        elif handler_call_details.method == _STREAM_UNARY:
            return _MethodHandler(
                True,
                False,
                _DESERIALIZE_REQUEST,
                _SERIALIZE_RESPONSE,
                None,
                None,
                self._handler.handle_stream_unary,
                None,
            )
        elif handler_call_details.method == _STREAM_STREAM:
            return _MethodHandler(
                True,
                True,
                None,
                None,
                None,
                None,
                None,
                self._handler.handle_stream_stream,
            )
        elif handler_call_details.method == _DEFECTIVE_GENERIC_RPC_HANDLER:
            return self._handler.defective_generic_rpc_handler()
        elif handler_call_details.method == _UNARY_UNARY_NESTED_EXCEPTION:
            return _MethodHandler(
                False,
                False,
                None,
                None,
                self._handler.handle_unary_unary_with_nested_exception,
                None,
                None,
                None,
            )
        else:
            return None


class FailAfterFewIterationsCounter(object):
    def __init__(self, high, bytestring):
        self._current = 0
        self._high = high
        self._bytestring = bytestring

    def __iter__(self):
        return self

    def __next__(self):
        if self._current >= self._high:
            raise test_control.Defect()
        else:
            self._current += 1
            return self._bytestring

    next = __next__


def _unary_unary_multi_callable(channel):
    return channel.unary_unary(_UNARY_UNARY)


def _unary_stream_multi_callable(channel):
    return channel.unary_stream(
        _UNARY_STREAM,
        request_serializer=_SERIALIZE_REQUEST,
        response_deserializer=_DESERIALIZE_RESPONSE,
    )


def _stream_unary_multi_callable(channel):
    return channel.stream_unary(
        _STREAM_UNARY,
        request_serializer=_SERIALIZE_REQUEST,
        response_deserializer=_DESERIALIZE_RESPONSE,
    )


def _stream_stream_multi_callable(channel):
    return channel.stream_stream(_STREAM_STREAM)


def _defective_handler_multi_callable(channel):
    return channel.unary_unary(_DEFECTIVE_GENERIC_RPC_HANDLER)


def _defective_nested_exception_handler_multi_callable(channel):
    return channel.unary_unary(_UNARY_UNARY_NESTED_EXCEPTION)


class InvocationDefectsTest(unittest.TestCase):
    """Tests the handling of exception-raising user code on the client-side."""

    def setUp(self):
        self._control = test_control.PauseFailControl()
        self._handler = _Handler(self._control)

        self._server = test_common.test_server()
        port = self._server.add_insecure_port("[::]:0")
        self._server.add_generic_rpc_handlers((_GenericHandler(self._handler),))
        self._server.start()

        self._channel = grpc.insecure_channel("localhost:%d" % port)

    def tearDown(self):
        self._server.stop(0)
        self._channel.close()

    def testIterableStreamRequestBlockingUnaryResponse(self):
        requests = object()
        multi_callable = _stream_unary_multi_callable(self._channel)

        with self.assertRaises(grpc.RpcError) as exception_context:
            multi_callable(
                requests,
                metadata=(
                    ("test", "IterableStreamRequestBlockingUnaryResponse"),
                ),
            )

        self.assertIs(
            grpc.StatusCode.UNKNOWN, exception_context.exception.code()
        )

    def testIterableStreamRequestFutureUnaryResponse(self):
        requests = object()
        multi_callable = _stream_unary_multi_callable(self._channel)
        response_future = multi_callable.future(
            requests,
            metadata=(("test", "IterableStreamRequestFutureUnaryResponse"),),
        )

        with self.assertRaises(grpc.RpcError) as exception_context:
            response_future.result()

        self.assertIs(
            grpc.StatusCode.UNKNOWN, exception_context.exception.code()
        )

    def testIterableStreamRequestStreamResponse(self):
        requests = object()
        multi_callable = _stream_stream_multi_callable(self._channel)
        response_iterator = multi_callable(
            requests,
            metadata=(("test", "IterableStreamRequestStreamResponse"),),
        )

        with self.assertRaises(grpc.RpcError) as exception_context:
            next(response_iterator)

        self.assertIs(
            grpc.StatusCode.UNKNOWN, exception_context.exception.code()
        )

    def testIteratorStreamRequestStreamResponse(self):
        requests_iterator = FailAfterFewIterationsCounter(
            test_constants.STREAM_LENGTH // 2, b"\x07\x08"
        )
        multi_callable = _stream_stream_multi_callable(self._channel)
        response_iterator = multi_callable(
            requests_iterator,
            metadata=(("test", "IteratorStreamRequestStreamResponse"),),
        )

        with self.assertRaises(grpc.RpcError) as exception_context:
            for _ in range(test_constants.STREAM_LENGTH // 2 + 1):
                next(response_iterator)

        self.assertIs(
            grpc.StatusCode.UNKNOWN, exception_context.exception.code()
        )

    def testDefectiveGenericRpcHandlerUnaryResponse(self):
        request = b"\x07\x08"
        multi_callable = _defective_handler_multi_callable(self._channel)

        with self.assertRaises(grpc.RpcError) as exception_context:
            multi_callable(
                request, metadata=(("test", "DefectiveGenericRpcHandlerUnary"),)
            )

        self.assertIs(
            grpc.StatusCode.UNKNOWN, exception_context.exception.code()
        )

    def testNestedExceptionGenericRpcHandlerUnaryResponse(self):
        request = b"\x07\x08"
        multi_callable = _defective_nested_exception_handler_multi_callable(
            self._channel
        )

        with self.assertRaises(grpc.RpcError) as exception_context:
            multi_callable(
                request, metadata=(("test", "DefectiveGenericRpcHandlerUnary"),)
            )

        self.assertIs(
            grpc.StatusCode.UNKNOWN, exception_context.exception.code()
        )


if __name__ == "__main__":
    logging.basicConfig()
    unittest.main(verbosity=2)



#!/usr/bin/python3

import sys
import os
import subprocess

srcdir = sys.argv[1]
outdir = sys.argv[2]

print("Building grpc s:%s, t:%s" % (srcdir, outdir), flush=True)

# Modifications to grpc tarball (1.59.0) #################################################################

# grpc/src/python/grpcio/support.py --- setuptools.errors.CompileError does not exist in python < 3.10.
#                                       replaced with distutils.errors.CompileError.
##########################################################################################################

try:
    os.chdir(srcdir)
    os.environ["GRPC_PYTHON_BUILD_EXT_COMPILER_JOBS"] = "2"
    subprocess.run(["python3", "setup.py", "build"])
    subprocess.run("cp -r python_build/lib*/grpc %s" % outdir, shell=True)
except Exception as e:
    print(e)
    sys.exit(1)

sys.exit(0)


# Copyright 2020 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides an interface to xDS Test Client running remotely.
"""
import datetime
import functools
import logging
from typing import Iterable, List, Optional

import framework.errors
from framework.helpers import retryers
import framework.rpc
from framework.rpc import grpc_channelz
from framework.rpc import grpc_csds
from framework.rpc import grpc_testing

logger = logging.getLogger(__name__)

# Type aliases
_timedelta = datetime.timedelta
_LoadBalancerStatsServiceClient = grpc_testing.LoadBalancerStatsServiceClient
_XdsUpdateClientConfigureServiceClient = (
    grpc_testing.XdsUpdateClientConfigureServiceClient
)
_ChannelzServiceClient = grpc_channelz.ChannelzServiceClient
_ChannelzChannel = grpc_channelz.Channel
_ChannelzChannelState = grpc_channelz.ChannelState
_ChannelzSubchannel = grpc_channelz.Subchannel
_ChannelzSocket = grpc_channelz.Socket
_CsdsClient = grpc_csds.CsdsClient

# Use in get_load_balancer_stats request to request all metadata.
REQ_LB_STATS_METADATA_ALL = ("*",)


class XdsTestClient(framework.rpc.grpc.GrpcApp):
    """
    Represents RPC services implemented in Client component of the xds test app.
    https://github.com/grpc/grpc/blob/master/doc/xds-test-descriptions.md#client
    """

    # A unique string identifying each client replica. Used in logging.
    hostname: str

    def __init__(
        self,
        *,
        ip: str,
        rpc_port: int,
        server_target: str,
        hostname: str,
        rpc_host: Optional[str] = None,
        maintenance_port: Optional[int] = None,
    ):
        super().__init__(rpc_host=(rpc_host or ip))
        self.ip = ip
        self.rpc_port = rpc_port
        self.server_target = server_target
        self.maintenance_port = maintenance_port or rpc_port
        self.hostname = hostname

    @property
    @functools.lru_cache(None)
    def load_balancer_stats(self) -> _LoadBalancerStatsServiceClient:
        return _LoadBalancerStatsServiceClient(
            self._make_channel(self.rpc_port),
            log_target=f"{self.hostname}:{self.rpc_port}",
        )

    @property
    @functools.lru_cache(None)
    def update_config(self):
        return _XdsUpdateClientConfigureServiceClient(
            self._make_channel(self.rpc_port),
            log_target=f"{self.hostname}:{self.rpc_port}",
        )

    @property
    @functools.lru_cache(None)
    def channelz(self) -> _ChannelzServiceClient:
        return _ChannelzServiceClient(
            self._make_channel(self.maintenance_port),
            log_target=f"{self.hostname}:{self.maintenance_port}",
        )

    @property
    @functools.lru_cache(None)
    def csds(self) -> _CsdsClient:
        return _CsdsClient(
            self._make_channel(self.maintenance_port),
            log_target=f"{self.hostname}:{self.maintenance_port}",
        )

    def get_load_balancer_stats(
        self,
        *,
        num_rpcs: int,
        metadata_keys: Optional[tuple[str, ...]] = None,
        timeout_sec: Optional[int] = None,
    ) -> grpc_testing.LoadBalancerStatsResponse:
        """
        Shortcut to LoadBalancerStatsServiceClient.get_client_stats()
        """
        return self.load_balancer_stats.get_client_stats(
            num_rpcs=num_rpcs,
            timeout_sec=timeout_sec,
            metadata_keys=metadata_keys,
        )

    def get_load_balancer_accumulated_stats(
        self,
        *,
        timeout_sec: Optional[int] = None,
    ) -> grpc_testing.LoadBalancerAccumulatedStatsResponse:
        """Shortcut to LoadBalancerStatsServiceClient.get_client_accumulated_stats()"""
        return self.load_balancer_stats.get_client_accumulated_stats(
            timeout_sec=timeout_sec
        )

    def wait_for_active_server_channel(self) -> _ChannelzChannel:
        """Wait for the channel to the server to transition to READY.

        Raises:
            GrpcApp.NotFound: If the channel never transitioned to READY.
        """
        try:
            return self.wait_for_server_channel_state(
                _ChannelzChannelState.READY
            )
        except retryers.RetryError as retry_err:
            if isinstance(retry_err.exception(), self.ChannelNotFound):
                retry_err.add_note(
                    framework.errors.FrameworkError.note_blanket_error(
                        "The client couldn't connect to the server."
                    )
                )
            raise

    def get_active_server_channel_socket(self) -> _ChannelzSocket:
        channel = self.find_server_channel_with_state(
            _ChannelzChannelState.READY
        )
        # Get the first subchannel of the active channel to the server.
        logger.debug(
            (
                "[%s] Retrieving client -> server socket, "
                "channel_id: %s, subchannel: %s"
            ),
            self.hostname,
            channel.ref.channel_id,
            channel.subchannel_ref[0].name,
        )
        subchannel, *subchannels = list(
            self.channelz.list_channel_subchannels(channel)
        )
        if subchannels:
            logger.warning(
                "[%s] Unexpected subchannels: %r", self.hostname, subchannels
            )
        # Get the first socket of the subchannel
        socket, *sockets = list(
            self.channelz.list_subchannels_sockets(subchannel)
        )
        if sockets:
            logger.warning(
                "[%s] Unexpected sockets: %r", self.hostname, subchannels
            )
        logger.debug(
            "[%s] Found client -> server socket: %s",
            self.hostname,
            socket.ref.name,
        )
        return socket

    def wait_for_server_channel_state(
        self,
        state: _ChannelzChannelState,
        *,
        timeout: Optional[_timedelta] = None,
        rpc_deadline: Optional[_timedelta] = None,
    ) -> _ChannelzChannel:
        # When polling for a state, prefer smaller wait times to avoid
        # exhausting all allowed time on a single long RPC.
        if rpc_deadline is None:
            rpc_deadline = _timedelta(seconds=30)

        # Fine-tuned to wait for the channel to the server.
        retryer = retryers.exponential_retryer_with_timeout(
            wait_min=_timedelta(seconds=10),
            wait_max=_timedelta(seconds=25),
            timeout=_timedelta(minutes=5) if timeout is None else timeout,
        )

        logger.info(
            "[%s] Waiting to report a %s channel to %s",
            self.hostname,
            _ChannelzChannelState.Name(state),
            self.server_target,
        )
        channel = retryer(
            self.find_server_channel_with_state,
            state,
            rpc_deadline=rpc_deadline,
        )
        logger.info(
            "[%s] Channel to %s transitioned to state %s: %s",
            self.hostname,
            self.server_target,
            _ChannelzChannelState.Name(state),
            _ChannelzServiceClient.channel_repr(channel),
        )
        return channel

    def find_server_channel_with_state(
        self,
        expected_state: _ChannelzChannelState,
        *,
        rpc_deadline: Optional[_timedelta] = None,
        check_subchannel=True,
    ) -> _ChannelzChannel:
        rpc_params = {}
        if rpc_deadline is not None:
            rpc_params["deadline_sec"] = rpc_deadline.total_seconds()

        expected_state_name: str = _ChannelzChannelState.Name(expected_state)
        target: str = self.server_target

        for channel in self.get_server_channels(target, **rpc_params):
            channel_state: _ChannelzChannelState = channel.data.state.state
            logger.info(
                "[%s] Server channel: %s",
                self.hostname,
                _ChannelzServiceClient.channel_repr(channel),
            )
            if channel_state is expected_state:
                if check_subchannel:
                    # When requested, check if the channel has at least
                    # one subchannel in the requested state.
                    try:
                        subchannel = self.find_subchannel_with_state(
                            channel, expected_state, **rpc_params
                        )
                        logger.info(
                            "[%s] Found subchannel in state %s: %s",
                            self.hostname,
                            expected_state_name,
                            _ChannelzServiceClient.subchannel_repr(subchannel),
                        )
                    except self.NotFound as e:
                        # Otherwise, keep searching.
                        logger.info(e.message)
                        continue
                return channel

        raise self.ChannelNotFound(
            f"[{self.hostname}] Client has no"
            f" {expected_state_name} channel with server {target}",
            src=self.hostname,
            dst=target,
            expected_state=expected_state,
        )

    def get_server_channels(
        self, server_target: str, **kwargs
    ) -> Iterable[_ChannelzChannel]:
        return self.channelz.find_channels_for_target(server_target, **kwargs)

    def find_subchannel_with_state(
        self, channel: _ChannelzChannel, state: _ChannelzChannelState, **kwargs
    ) -> _ChannelzSubchannel:
        subchannels = self.channelz.list_channel_subchannels(channel, **kwargs)
        for subchannel in subchannels:
            if subchannel.data.state.state is state:
                return subchannel

        raise self.NotFound(
            f"[{self.hostname}] Not found "
            f"a {_ChannelzChannelState.Name(state)} subchannel "
            f"for channel_id {channel.ref.channel_id}"
        )

    def find_subchannels_with_state(
        self, state: _ChannelzChannelState, **kwargs
    ) -> List[_ChannelzSubchannel]:
        subchannels = []
        for channel in self.channelz.find_channels_for_target(
            self.server_target, **kwargs
        ):
            for subchannel in self.channelz.list_channel_subchannels(
                channel, **kwargs
            ):
                if subchannel.data.state.state is state:
                    subchannels.append(subchannel)
        return subchannels

    class ChannelNotFound(framework.rpc.grpc.GrpcApp.NotFound):
        """Channel with expected status not found"""

        src: str
        dst: str
        expected_state: object

        def __init__(
            self,
            message: str,
            *,
            src: str,
            dst: str,
            expected_state: _ChannelzChannelState,
            **kwargs,
        ):
            self.src = src
            self.dst = dst
            self.expected_state = expected_state
            super().__init__(message, src, dst, expected_state, **kwargs)



#!/usr/bin/env python3
# Copyright 2023 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.
"""Single certificate chain for serverAuth which is self-signed."""

import sys
sys.path += ['../..']

import gencerts

# Self-signed target certificate.
target = gencerts.create_self_signed_end_entity_certificate('Target')
target.get_extensions().set_property('extendedKeyUsage', 'serverAuth')

chain = [target]
gencerts.write_chain(__doc__, chain, 'chain.pem')



# Copyright 2022 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import logging

from absl import flags
from absl.testing import absltest
import grpc

from framework import xds_k8s_flags
from framework import xds_k8s_testcase
from framework.helpers import skips

logger = logging.getLogger(__name__)
flags.adopt_module_key_flags(xds_k8s_testcase)

# Type aliases
_XdsTestServer = xds_k8s_testcase.XdsTestServer
_XdsTestClient = xds_k8s_testcase.XdsTestClient
_Lang = skips.Lang

_EXPECTED_STATUS = grpc.StatusCode.DATA_LOSS


class CustomLbTest(xds_k8s_testcase.RegularXdsKubernetesTestCase):
    @classmethod
    def setUpClass(cls):
        """Force the java test server for languages not yet supporting
        the `rpc-behavior` feature.
        https://github.com/grpc/grpc/blob/master/doc/xds-test-descriptions.md#server
        """
        super().setUpClass()
        # gRPC Java implemented server "error-code-" rpc-behavior in v1.47.x.
        # gRPC CPP implemented rpc-behavior in the same version, as custom_lb.
        if cls.lang_spec.client_lang in _Lang.JAVA | _Lang.CPP:
            return

        # gRPC go, python and node fallback to the gRPC Java.
        # TODO(https://github.com/grpc/grpc-go/issues/6288): use go server.
        # TODO(https://github.com/grpc/grpc/issues/33134): use python server.
        cls.server_image = xds_k8s_flags.SERVER_IMAGE_CANONICAL.value

    @staticmethod
    def is_supported(config: skips.TestConfig) -> bool:
        if config.client_lang == _Lang.JAVA:
            return config.version_gte("v1.47.x")
        if config.client_lang == _Lang.CPP:
            return config.version_gte("v1.55.x")
        if config.client_lang == _Lang.GO:
            return config.version_gte("v1.56.x")
        if config.client_lang == _Lang.NODE:
            return config.version_gte("v1.10.x")
        return False

    def test_custom_lb_config(self):
        with self.subTest("0_create_health_check"):
            self.td.create_health_check()

        # Configures a custom, test LB on the client to instruct the servers
        # to always respond with a specific error code.
        #
        # The first policy in the list is a non-existent one to verify that
        # the gRPC client can gracefully move down the list to the valid one
        # once it determines the first one is not available.
        with self.subTest("1_create_backend_service"):
            self.td.create_backend_service(
                locality_lb_policies=[
                    {
                        "customPolicy": {
                            "name": "test.ThisLoadBalancerDoesNotExist",
                            "data": '{ "foo": "bar" }',
                        },
                    },
                    {
                        "customPolicy": {
                            "name": "test.RpcBehaviorLoadBalancer",
                            "data": (
                                '{ "rpcBehavior":'
                                f' "error-code-{_EXPECTED_STATUS.value[0]}" }}'
                            ),
                        }
                    },
                ]
            )

        with self.subTest("2_create_url_map"):
            self.td.create_url_map(self.server_xds_host, self.server_xds_port)

        with self.subTest("3_create_target_proxy"):
            self.td.create_target_proxy()

        with self.subTest("4_create_forwarding_rule"):
            self.td.create_forwarding_rule(self.server_xds_port)

        with self.subTest("5_start_test_server"):
            test_server: _XdsTestServer = self.startTestServers()[0]

        with self.subTest("6_add_server_backends_to_backend_service"):
            self.setupServerBackends()

        with self.subTest("7_start_test_client"):
            test_client: _XdsTestClient = self.startTestClient(test_server)

        with self.subTest("8_test_client_xds_config_exists"):
            self.assertXdsConfigExists(test_client)

        # Verify status codes from the servers have the configured one.
        with self.subTest("9_test_server_returned_configured_status_code"):
            self.assertRpcStatusCodes(
                test_client,
                expected_status=_EXPECTED_STATUS,
                duration=datetime.timedelta(seconds=10),
                method="UNARY_CALL",
            )


if __name__ == "__main__":
    absltest.main(failfast=True)



""" Multicast DNS Service Discovery for Python, v0.14-wmcbrine
    Copyright 2003 Paul Scott-Murphy, 2014 William McBrine

    This module provides a framework for the use of DNS Service Discovery
    using IP multicast.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
    USA
"""
import asyncio
import contextlib
from types import TracebackType  # noqa # used in type hints
from typing import Awaitable, Callable, Dict, List, Optional, Tuple, Type, Union

from ._core import Zeroconf
from ._dns import DNSQuestionType
from ._services import ServiceListener
from ._services.browser import _ServiceBrowserBase
from ._services.info import ServiceInfo
from ._services.types import ZeroconfServiceTypes
from ._utils.net import InterfaceChoice, InterfacesType, IPVersion
from .const import _BROWSER_TIME, _MDNS_PORT, _SERVICE_TYPE_ENUMERATION_NAME

__all__ = [
    "AsyncZeroconf",
    "AsyncServiceInfo",
    "AsyncServiceBrowser",
    "AsyncZeroconfServiceTypes",
]


class AsyncServiceInfo(ServiceInfo):
    """An async version of ServiceInfo."""


class AsyncServiceBrowser(_ServiceBrowserBase):
    """Used to browse for a service for specific type(s).

    Constructor parameters are as follows:

    * `zc`: A Zeroconf instance
    * `type_`: fully qualified service type name
    * `handler`: ServiceListener or Callable that knows how to process ServiceStateChange events
    * `listener`: ServiceListener
    * `addr`: address to send queries (will default to multicast)
    * `port`: port to send queries (will default to mdns 5353)
    * `delay`: The initial delay between answering questions
    * `question_type`: The type of questions to ask (DNSQuestionType.QM or DNSQuestionType.QU)

    The listener object will have its add_service() and
    remove_service() methods called when this browser
    discovers changes in the services availability.
    """

    def __init__(
        self,
        zeroconf: 'Zeroconf',
        type_: Union[str, list],
        handlers: Optional[Union[ServiceListener, List[Callable[..., None]]]] = None,
        listener: Optional[ServiceListener] = None,
        addr: Optional[str] = None,
        port: int = _MDNS_PORT,
        delay: int = _BROWSER_TIME,
        question_type: Optional[DNSQuestionType] = None,
    ) -> None:
        super().__init__(zeroconf, type_, handlers, listener, addr, port, delay, question_type)
        self._async_start()

    async def async_cancel(self) -> None:
        """Cancel the browser."""
        self._async_cancel()

    async def __aenter__(self) -> 'AsyncServiceBrowser':
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> Optional[bool]:
        await self.async_cancel()
        return None


class AsyncZeroconfServiceTypes(ZeroconfServiceTypes):
    """An async version of ZeroconfServiceTypes."""

    @classmethod
    async def async_find(
        cls,
        aiozc: Optional['AsyncZeroconf'] = None,
        timeout: Union[int, float] = 5,
        interfaces: InterfacesType = InterfaceChoice.All,
        ip_version: Optional[IPVersion] = None,
    ) -> Tuple[str, ...]:
        """
        Return all of the advertised services on any local networks.

        :param aiozc: AsyncZeroconf() instance.  Pass in if already have an
                instance running or if non-default interfaces are needed
        :param timeout: seconds to wait for any responses
        :param interfaces: interfaces to listen on.
        :param ip_version: IP protocol version to use.
        :return: tuple of service type strings
        """
        local_zc = aiozc or AsyncZeroconf(interfaces=interfaces, ip_version=ip_version)
        listener = cls()
        async_browser = AsyncServiceBrowser(
            local_zc.zeroconf, _SERVICE_TYPE_ENUMERATION_NAME, listener=listener
        )

        # wait for responses
        await asyncio.sleep(timeout)

        await async_browser.async_cancel()

        # close down anything we opened
        if aiozc is None:
            await local_zc.async_close()

        return tuple(sorted(listener.found_services))


class AsyncZeroconf:
    """Implementation of Zeroconf Multicast DNS Service Discovery

    Supports registration, unregistration, queries and browsing.

    The async version is currently a wrapper around Zeroconf which
    is now also async. It is expected that an asyncio event loop
    is already running before creating the AsyncZeroconf object.
    """

    def __init__(
        self,
        interfaces: InterfacesType = InterfaceChoice.All,
        unicast: bool = False,
        ip_version: Optional[IPVersion] = None,
        apple_p2p: bool = False,
        zc: Optional[Zeroconf] = None,
    ) -> None:
        """Creates an instance of the Zeroconf class, establishing
        multicast communications, and listening.

        :param interfaces: :class:`InterfaceChoice` or a list of IP addresses
            (IPv4 and IPv6) and interface indexes (IPv6 only).

            IPv6 notes for non-POSIX systems:
            * `InterfaceChoice.All` is an alias for `InterfaceChoice.Default`
              on Python versions before 3.8.

            Also listening on loopback (``::1``) doesn't work, use a real address.
        :param ip_version: IP versions to support. If `choice` is a list, the default is detected
            from it. Otherwise defaults to V4 only for backward compatibility.
        :param apple_p2p: use AWDL interface (only macOS)
        """
        self.zeroconf = zc or Zeroconf(
            interfaces=interfaces,
            unicast=unicast,
            ip_version=ip_version,
            apple_p2p=apple_p2p,
        )
        self.async_browsers: Dict[ServiceListener, AsyncServiceBrowser] = {}

    async def async_register_service(
        self,
        info: ServiceInfo,
        ttl: Optional[int] = None,
        allow_name_change: bool = False,
        cooperating_responders: bool = False,
        strict: bool = True,
    ) -> Awaitable:
        """Registers service information to the network with a default TTL.
        Zeroconf will then respond to requests for information for that
        service.  The name of the service may be changed if needed to make
        it unique on the network. Additionally multiple cooperating responders
        can register the same service on the network for resilience
        (if you want this behavior set `cooperating_responders` to `True`).

        The service will be broadcast in a task. This task is returned
        and therefore can be awaited if necessary.
        """
        return await self.zeroconf.async_register_service(
            info, ttl, allow_name_change, cooperating_responders, strict
        )

    async def async_unregister_all_services(self) -> None:
        """Unregister all registered services.

        Unlike async_register_service and async_unregister_service, this
        method does not return a future and is always expected to be
        awaited since its only called at shutdown.
        """
        await self.zeroconf.async_unregister_all_services()

    async def async_unregister_service(self, info: ServiceInfo) -> Awaitable:
        """Unregister a service.

        The service will be broadcast in a task. This task is returned
        and therefore can be awaited if necessary.
        """
        return await self.zeroconf.async_unregister_service(info)

    async def async_update_service(self, info: ServiceInfo) -> Awaitable:
        """Registers service information to the network with a default TTL.
        Zeroconf will then respond to requests for information for that
        service.

        The service will be broadcast in a task. This task is returned
        and therefore can be awaited if necessary.
        """
        return await self.zeroconf.async_update_service(info)

    async def async_close(self) -> None:
        """Ends the background threads, and prevent this instance from
        servicing further queries."""
        if not self.zeroconf.done:
            with contextlib.suppress(asyncio.TimeoutError):
                await asyncio.wait_for(self.zeroconf.async_wait_for_start(), timeout=1)
        await self.async_remove_all_service_listeners()
        await self.async_unregister_all_services()
        await self.zeroconf._async_close()  # pylint: disable=protected-access

    async def async_get_service_info(
        self, type_: str, name: str, timeout: int = 3000, question_type: Optional[DNSQuestionType] = None
    ) -> Optional[AsyncServiceInfo]:
        """Returns network's service information for a particular
        name and type, or None if no service matches by the timeout,
        which defaults to 3 seconds."""
        info = AsyncServiceInfo(type_, name)
        if await info.async_request(self.zeroconf, timeout, question_type):
            return info
        return None

    async def async_add_service_listener(self, type_: str, listener: ServiceListener) -> None:
        """Adds a listener for a particular service type.  This object
        will then have its add_service and remove_service methods called when
        services of that type become available and unavailable."""
        await self.async_remove_service_listener(listener)
        self.async_browsers[listener] = AsyncServiceBrowser(self.zeroconf, type_, listener)

    async def async_remove_service_listener(self, listener: ServiceListener) -> None:
        """Removes a listener from the set that is currently listening."""
        if listener in self.async_browsers:
            await self.async_browsers[listener].async_cancel()
            del self.async_browsers[listener]

    async def async_remove_all_service_listeners(self) -> None:
        """Removes a listener from the set that is currently listening."""
        await asyncio.gather(
            *(self.async_remove_service_listener(listener) for listener in list(self.async_browsers))
        )

    async def __aenter__(self) -> 'AsyncZeroconf':
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> Optional[bool]:
        await self.async_close()
        return None



# Copyright 2020 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides an interface to xDS Test Server running remotely.
"""
import functools
import logging
from typing import Iterator, Optional

import framework.rpc
from framework.rpc import grpc_channelz
from framework.rpc import grpc_testing

logger = logging.getLogger(__name__)

# Type aliases
_ChannelzServiceClient = grpc_channelz.ChannelzServiceClient
_XdsUpdateHealthServiceClient = grpc_testing.XdsUpdateHealthServiceClient
_HealthClient = grpc_testing.HealthClient


class XdsTestServer(framework.rpc.grpc.GrpcApp):
    """
    Represents RPC services implemented in Server component of the xDS test app.
    https://github.com/grpc/grpc/blob/master/doc/xds-test-descriptions.md#server
    """

    # A unique host name identifying each server replica.
    # Server implementation must return this in the SimpleResponse.hostname,
    # which client uses as the key in rpcs_by_peer map.
    hostname: str

    def __init__(
        self,
        *,
        ip: str,
        rpc_port: int,
        hostname: str,
        maintenance_port: Optional[int] = None,
        secure_mode: Optional[bool] = False,
        xds_host: Optional[str] = None,
        xds_port: Optional[int] = None,
        rpc_host: Optional[str] = None,
    ):
        super().__init__(rpc_host=(rpc_host or ip))
        self.ip = ip
        self.rpc_port = rpc_port
        self.hostname = hostname
        self.maintenance_port = maintenance_port or rpc_port
        self.secure_mode = secure_mode
        self.xds_host, self.xds_port = xds_host, xds_port

    @property
    @functools.lru_cache(None)
    def channelz(self) -> _ChannelzServiceClient:
        return _ChannelzServiceClient(
            self._make_channel(self.maintenance_port),
            log_target=f"{self.hostname}:{self.maintenance_port}",
        )

    @property
    @functools.lru_cache(None)
    def update_health_service_client(self) -> _XdsUpdateHealthServiceClient:
        return _XdsUpdateHealthServiceClient(
            self._make_channel(self.maintenance_port),
            log_target=f"{self.hostname}:{self.maintenance_port}",
        )

    @property
    @functools.lru_cache(None)
    def health_client(self) -> _HealthClient:
        return _HealthClient(
            self._make_channel(self.maintenance_port),
            log_target=f"{self.hostname}:{self.maintenance_port}",
        )

    def set_serving(self):
        logger.info("[%s] >> Setting health status to SERVING", self.hostname)
        self.update_health_service_client.set_serving()
        logger.info(
            "[%s] << Health status %s",
            self.hostname,
            self.health_client.check_health(),
        )

    def set_not_serving(self):
        logger.info(
            "[%s] >> Setting health status to NOT_SERVING", self.hostname
        )
        self.update_health_service_client.set_not_serving()
        logger.info(
            "[%s] << Health status %s",
            self.hostname,
            self.health_client.check_health(),
        )

    def set_xds_address(self, xds_host, xds_port: Optional[int] = None):
        self.xds_host, self.xds_port = xds_host, xds_port

    @property
    def xds_address(self) -> str:
        if not self.xds_host:
            return ""
        if not self.xds_port:
            return self.xds_host
        return f"{self.xds_host}:{self.xds_port}"

    @property
    def xds_uri(self) -> str:
        if not self.xds_host:
            return ""
        return f"xds:///{self.xds_address}"

    def get_test_server(self) -> grpc_channelz.Server:
        """Return channelz representation of a server running TestService.

        Raises:
            GrpcApp.NotFound: Test server not found.
        """
        server = self.channelz.find_server_listening_on_port(self.rpc_port)
        if not server:
            raise self.NotFound(
                f"[{self.hostname}] Server"
                f"listening on port {self.rpc_port} not found"
            )
        return server

    def get_test_server_sockets(self) -> Iterator[grpc_channelz.Socket]:
        """List all sockets of the test server.

        Raises:
            GrpcApp.NotFound: Test server not found.
        """
        server = self.get_test_server()
        return self.channelz.list_server_sockets(server)

    def get_server_socket_matching_client(
        self, client_socket: grpc_channelz.Socket
    ):
        """Find test server socket that matches given test client socket.

        Sockets are matched using TCP endpoints (ip:port), further on "address".
        Server socket remote address matched with client socket local address.

         Raises:
             GrpcApp.NotFound: Server socket matching client socket not found.
        """
        client_local = self.channelz.sock_address_to_str(client_socket.local)
        logger.debug(
            "[%s] Looking for a server socket connected to the client %s",
            self.hostname,
            client_local,
        )

        server_socket = self.channelz.find_server_socket_matching_client(
            self.get_test_server_sockets(), client_socket
        )
        if not server_socket:
            raise self.NotFound(
                f"[{self.hostname}] Socket to client {client_local} not found"
            )

        logger.info(
            "[%s] Found matching socket pair: server(%s) <-> client(%s)",
            self.hostname,
            self.channelz.sock_addresses_pretty(server_socket),
            self.channelz.sock_addresses_pretty(client_socket),
        )
        return server_socket



# Copyright 2016 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import threading

from src.proto.grpc.testing import stats_pb2


class Histogram(object):
    """Histogram class used for recording performance testing data.

    This class is thread safe.
    """

    def __init__(self, resolution, max_possible):
        self._lock = threading.Lock()
        self._resolution = resolution
        self._max_possible = max_possible
        self._sum = 0
        self._sum_of_squares = 0
        self.multiplier = 1.0 + self._resolution
        self._count = 0
        self._min = self._max_possible
        self._max = 0
        self._buckets = [0] * (self._bucket_for(self._max_possible) + 1)

    def reset(self):
        with self._lock:
            self._sum = 0
            self._sum_of_squares = 0
            self._count = 0
            self._min = self._max_possible
            self._max = 0
            self._buckets = [0] * (self._bucket_for(self._max_possible) + 1)

    def add(self, val):
        with self._lock:
            self._sum += val
            self._sum_of_squares += val * val
            self._count += 1
            self._min = min(self._min, val)
            self._max = max(self._max, val)
            self._buckets[self._bucket_for(val)] += 1

    def get_data(self):
        with self._lock:
            data = stats_pb2.HistogramData()
            data.bucket.extend(self._buckets)
            data.min_seen = self._min
            data.max_seen = self._max
            data.sum = self._sum
            data.sum_of_squares = self._sum_of_squares
            data.count = self._count
            return data

    def merge(self, another_data):
        with self._lock:
            for i in range(len(self._buckets)):
                self._buckets[i] += another_data.bucket[i]
            self._min = min(self._min, another_data.min_seen)
            self._max = max(self._max, another_data.max_seen)
            self._sum += another_data.sum
            self._sum_of_squares += another_data.sum_of_squares
            self._count += another_data.count

    def _bucket_for(self, val):
        val = min(val, self._max_possible)
        return int(math.log(val, self.multiplier))



#!/usr/bin/env python


"""Unit tests for zeroconf._services.types."""

import logging
import os
import socket
import sys
import unittest

import zeroconf as r
from zeroconf import ServiceInfo, Zeroconf, ZeroconfServiceTypes

from .. import _clear_cache, has_working_ipv6

log = logging.getLogger('zeroconf')
original_logging_level = logging.NOTSET


def setup_module():
    global original_logging_level
    original_logging_level = log.level
    log.setLevel(logging.DEBUG)


def teardown_module():
    if original_logging_level != logging.NOTSET:
        log.setLevel(original_logging_level)


def test_integration_with_listener(disable_duplicate_packet_suppression):
    type_ = "_test-listen-type._tcp.local."
    name = "xxxyyy"
    registration_name = f"{name}.{type_}"

    zeroconf_registrar = Zeroconf(interfaces=['127.0.0.1'])
    desc = {'path': '/~paulsm/'}
    info = ServiceInfo(
        type_,
        registration_name,
        80,
        0,
        0,
        desc,
        "ash-2.local.",
        addresses=[socket.inet_aton("10.0.1.2")],
    )
    zeroconf_registrar.registry.async_add(info)
    try:
        service_types = ZeroconfServiceTypes.find(interfaces=['127.0.0.1'], timeout=2)
        assert type_ in service_types
        _clear_cache(zeroconf_registrar)
        service_types = ZeroconfServiceTypes.find(zc=zeroconf_registrar, timeout=2)
        assert type_ in service_types

    finally:
        zeroconf_registrar.close()


@unittest.skipIf(not has_working_ipv6(), 'Requires IPv6')
@unittest.skipIf(os.environ.get('SKIP_IPV6'), 'IPv6 tests disabled')
def test_integration_with_listener_v6_records(disable_duplicate_packet_suppression):
    type_ = "_test-listenv6rec-type._tcp.local."
    name = "xxxyyy"
    registration_name = f"{name}.{type_}"
    addr = "2606:2800:220:1:248:1893:25c8:1946"  # example.com

    zeroconf_registrar = Zeroconf(interfaces=['127.0.0.1'])
    desc = {'path': '/~paulsm/'}
    info = ServiceInfo(
        type_,
        registration_name,
        80,
        0,
        0,
        desc,
        "ash-2.local.",
        addresses=[socket.inet_pton(socket.AF_INET6, addr)],
    )
    zeroconf_registrar.registry.async_add(info)
    try:
        service_types = ZeroconfServiceTypes.find(interfaces=['127.0.0.1'], timeout=2)
        assert type_ in service_types
        _clear_cache(zeroconf_registrar)
        service_types = ZeroconfServiceTypes.find(zc=zeroconf_registrar, timeout=2)
        assert type_ in service_types

    finally:
        zeroconf_registrar.close()


@unittest.skipIf(not has_working_ipv6() or sys.platform == 'win32', 'Requires IPv6')
@unittest.skipIf(os.environ.get('SKIP_IPV6'), 'IPv6 tests disabled')
def test_integration_with_listener_ipv6(disable_duplicate_packet_suppression):
    type_ = "_test-listenv6ip-type._tcp.local."
    name = "xxxyyy"
    registration_name = f"{name}.{type_}"
    addr = "2606:2800:220:1:248:1893:25c8:1946"  # example.com

    zeroconf_registrar = Zeroconf(ip_version=r.IPVersion.V6Only)
    desc = {'path': '/~paulsm/'}
    info = ServiceInfo(
        type_,
        registration_name,
        80,
        0,
        0,
        desc,
        "ash-2.local.",
        addresses=[socket.inet_pton(socket.AF_INET6, addr)],
    )
    zeroconf_registrar.registry.async_add(info)
    try:
        service_types = ZeroconfServiceTypes.find(ip_version=r.IPVersion.V6Only, timeout=2)
        assert type_ in service_types
        _clear_cache(zeroconf_registrar)
        service_types = ZeroconfServiceTypes.find(zc=zeroconf_registrar, timeout=2)
        assert type_ in service_types

    finally:
        zeroconf_registrar.close()


def test_integration_with_subtype_and_listener(disable_duplicate_packet_suppression):
    subtype_ = "_subtype._sub"
    type_ = "_listen._tcp.local."
    name = "xxxyyy"
    # Note: discovery returns only DNS-SD type not subtype
    discovery_type = f"{subtype_}.{type_}"
    registration_name = f"{name}.{type_}"

    zeroconf_registrar = Zeroconf(interfaces=['127.0.0.1'])
    desc = {'path': '/~paulsm/'}
    info = ServiceInfo(
        discovery_type,
        registration_name,
        80,
        0,
        0,
        desc,
        "ash-2.local.",
        addresses=[socket.inet_aton("10.0.1.2")],
    )
    zeroconf_registrar.registry.async_add(info)
    try:
        service_types = ZeroconfServiceTypes.find(interfaces=['127.0.0.1'], timeout=2)
        assert discovery_type in service_types
        _clear_cache(zeroconf_registrar)
        service_types = ZeroconfServiceTypes.find(zc=zeroconf_registrar, timeout=2)
        assert discovery_type in service_types

    finally:
        zeroconf_registrar.close()



# Copyright 2015 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests of grpc.beta._connectivity_channel."""

import unittest

from grpc.beta import interfaces


class ConnectivityStatesTest(unittest.TestCase):
    def testBetaConnectivityStates(self):
        self.assertIsNotNone(interfaces.ChannelConnectivity.IDLE)
        self.assertIsNotNone(interfaces.ChannelConnectivity.CONNECTING)
        self.assertIsNotNone(interfaces.ChannelConnectivity.READY)
        self.assertIsNotNone(interfaces.ChannelConnectivity.TRANSIENT_FAILURE)
        self.assertIsNotNone(interfaces.ChannelConnectivity.FATAL_FAILURE)


if __name__ == "__main__":
    unittest.main(verbosity=2)



# Copyright 2023 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from typing import List, Optional

from absl import flags
from absl.testing import absltest

from framework import xds_gamma_testcase
from framework import xds_k8s_testcase
from framework import xds_url_map_testcase
from framework.rpc import grpc_testing
from framework.test_app import client_app
from framework.test_app import server_app
from framework.test_cases import session_affinity_util

logger = logging.getLogger(__name__)
flags.adopt_module_key_flags(xds_k8s_testcase)

_XdsTestServer = server_app.XdsTestServer
_XdsTestClient = client_app.XdsTestClient
RpcTypeUnaryCall = xds_url_map_testcase.RpcTypeUnaryCall

_REPLICA_COUNT = 3


class AffinityTest(xds_gamma_testcase.GammaXdsKubernetesTestCase):
    def getClientRpcStats(
        self,
        test_client: _XdsTestClient,
        num_rpcs: int,
        *,
        metadata_keys: Optional[tuple[str, ...]] = None,
    ) -> grpc_testing.LoadBalancerStatsResponse:
        """Load all metadata_keys by default."""
        return super().getClientRpcStats(
            test_client,
            num_rpcs,
            metadata_keys=metadata_keys or client_app.REQ_LB_STATS_METADATA_ALL,
        )

    def test_session_affinity_filter(self):
        test_servers: List[_XdsTestServer]
        with self.subTest("01_run_test_server"):
            test_servers = self.startTestServers(
                replica_count=_REPLICA_COUNT,
                route_template="gamma/route_http_ssafilter.yaml",
            )

        with self.subTest("02_create_ssa_filter"):
            self.server_runner.createSessionAffinityFilter()

        # Default is round robin LB policy.

        with self.subTest("03_start_test_client"):
            test_client: _XdsTestClient = self.startTestClient(test_servers[0])

        with self.subTest("04_send_first_RPC_and_retrieve_cookie"):
            (
                cookie,
                chosen_server,
            ) = session_affinity_util.assert_eventually_retrieve_cookie_and_server(
                self, test_client, test_servers
            )

        with self.subTest("05_send_RPCs_with_cookie"):
            test_client.update_config.configure(
                rpc_types=(RpcTypeUnaryCall,),
                metadata=(
                    (
                        RpcTypeUnaryCall,
                        "cookie",
                        cookie,
                    ),
                ),
            )
            self.assertRpcsEventuallyGoToGivenServers(
                test_client, [chosen_server], 10
            )

    def test_session_affinity_policy_with_route_target(self):
        test_servers: List[_XdsTestServer]
        with self.subTest("01_run_test_server"):
            test_servers = self.startTestServers(replica_count=_REPLICA_COUNT)

        with self.subTest("02_create_ssa_policy"):
            self.server_runner.createSessionAffinityPolicy(
                "gamma/session_affinity_policy_route.yaml"
            )

        # Default is round robin LB policy.

        with self.subTest("03_start_test_client"):
            test_client: _XdsTestClient = self.startTestClient(test_servers[0])

        with self.subTest("04_send_first_RPC_and_retrieve_cookie"):
            (
                cookie,
                chosen_server,
            ) = session_affinity_util.assert_eventually_retrieve_cookie_and_server(
                self, test_client, test_servers
            )

        with self.subTest("05_send_RPCs_with_cookie"):
            test_client.update_config.configure(
                rpc_types=(RpcTypeUnaryCall,),
                metadata=(
                    (
                        RpcTypeUnaryCall,
                        "cookie",
                        cookie,
                    ),
                ),
            )
            self.assertRpcsEventuallyGoToGivenServers(
                test_client, [chosen_server], 10
            )

    def test_session_affinity_policy_with_service_target(self):
        test_servers: List[_XdsTestServer]
        with self.subTest("01_run_test_server"):
            test_servers = self.startTestServers(replica_count=_REPLICA_COUNT)

        with self.subTest("02_create_ssa_policy"):
            self.server_runner.createSessionAffinityPolicy(
                "gamma/session_affinity_policy_service.yaml"
            )

        # Default is round robin LB policy.

        with self.subTest("03_start_test_client"):
            test_client: _XdsTestClient = self.startTestClient(test_servers[0])

        with self.subTest("04_send_first_RPC_and_retrieve_cookie"):
            (
                cookie,
                chosen_server,
            ) = session_affinity_util.assert_eventually_retrieve_cookie_and_server(
                self, test_client, test_servers
            )

        with self.subTest("05_send_RPCs_with_cookie"):
            test_client.update_config.configure(
                rpc_types=(RpcTypeUnaryCall,),
                metadata=(
                    (
                        RpcTypeUnaryCall,
                        "cookie",
                        cookie,
                    ),
                ),
            )
            self.assertRpcsEventuallyGoToGivenServers(
                test_client, [chosen_server], 10
            )


if __name__ == "__main__":
    absltest.main()



# Copyright 2023 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This contains common helpers for working with grpc data structures."""
import dataclasses
import functools
from typing import Dict, List, Optional

import grpc
import yaml

from framework.rpc import grpc_testing

# Type aliases
RpcsByPeer: Dict[str, int]
RpcMetadata = grpc_testing.LoadBalancerStatsResponse.RpcMetadata
MetadataByPeer: list[str, RpcMetadata]


@functools.cache  # pylint: disable=no-member
def status_from_int(grpc_status_int: int) -> Optional[grpc.StatusCode]:
    """Converts the integer gRPC status code to the grpc.StatusCode enum."""
    for grpc_status in grpc.StatusCode:
        if grpc_status.value[0] == grpc_status_int:
            return grpc_status
    return None


def status_eq(grpc_status_int: int, grpc_status: grpc.StatusCode) -> bool:
    """Compares the integer gRPC status code with the grpc.StatusCode enum."""
    return status_from_int(grpc_status_int) is grpc_status


def status_pretty(grpc_status: grpc.StatusCode) -> str:
    """Formats the status code as (int, NAME), f.e. (4, DEADLINE_EXCEEDED)"""
    return f"({grpc_status.value[0]}, {grpc_status.name})"


@dataclasses.dataclass(frozen=True)
class PrettyStatsPerMethod:
    # The name of the method.
    method: str

    # The number of RPCs started for this method, completed and in-flight.
    rpcs_started: int

    # The number of RPCs that completed with each status for this method.
    # Format: status code -> RPC count, f.e.:
    # {
    #   "(0, OK)": 20,
    #   "(14, UNAVAILABLE)": 10
    # }
    result: Dict[str, int]

    @functools.cached_property  # pylint: disable=no-member
    def rpcs_completed(self):
        """Returns the total count of competed RPCs across all statuses."""
        return sum(self.result.values())

    @staticmethod
    def from_response(
        method_name: str, method_stats: grpc_testing.MethodStats
    ) -> "PrettyStatsPerMethod":
        stats: Dict[str, int] = dict()
        for status_int, count in method_stats.result.items():
            status: Optional[grpc.StatusCode] = status_from_int(status_int)
            status_formatted = status_pretty(status) if status else "None"
            stats[status_formatted] = count
        return PrettyStatsPerMethod(
            method=method_name,
            rpcs_started=method_stats.rpcs_started,
            result=stats,
        )


def accumulated_stats_pretty(
    accumulated_stats: grpc_testing.LoadBalancerAccumulatedStatsResponse,
    *,
    ignore_empty: bool = False,
) -> str:
    """Pretty print LoadBalancerAccumulatedStatsResponse.

    Example:
      - method: EMPTY_CALL
        rpcs_started: 0
        result:
          (2, UNKNOWN): 20
      - method: UNARY_CALL
        rpcs_started: 31
        result:
          (0, OK): 10
          (14, UNAVAILABLE): 20
    """
    # Only look at stats_per_method, as the other fields are deprecated.
    result: List[Dict] = []
    for method_name, method_stats in accumulated_stats.stats_per_method.items():
        pretty_stats = PrettyStatsPerMethod.from_response(
            method_name, method_stats
        )
        # Skip methods with no RPCs reported when ignore_empty is True.
        if ignore_empty and not pretty_stats.rpcs_started:
            continue
        result.append(dataclasses.asdict(pretty_stats))

    return yaml.dump(result, sort_keys=False)


@dataclasses.dataclass(frozen=True)
class PrettyLoadBalancerStats:
    # The number of RPCs that failed to record a remote peer.
    num_failures: int

    # The number of completed RPCs for each peer.
    # Format: a dictionary from the host name (str) to the RPC count (int), f.e.
    # {"host-a": 10, "host-b": 20}
    rpcs_by_peer: "RpcsByPeer"

    # The number of completed RPCs per method per each pear.
    # Format: a dictionary from the method name to RpcsByPeer (see above), f.e.:
    # {
    #   "UNARY_CALL": {"host-a": 10, "host-b": 20},
    #   "EMPTY_CALL": {"host-a": 42},
    # }
    rpcs_by_method: Dict[str, "RpcsByPeer"]

    metadatas_by_peer: Dict[str, "MetadataByPeer"]

    @staticmethod
    def _parse_rpcs_by_peer(
        rpcs_by_peer: grpc_testing.RpcsByPeer,
    ) -> "RpcsByPeer":
        result = dict()
        for peer, count in rpcs_by_peer.items():
            result[peer] = count
        return result

    @staticmethod
    def _parse_metadatas_by_peer(
        metadatas_by_peer: grpc_testing.LoadBalancerStatsResponse.MetadataByPeer,
    ) -> "MetadataByPeer":
        result = dict()
        for peer, metadatas in metadatas_by_peer.items():
            pretty_metadata = ""
            for rpc_metadatas in metadatas.rpc_metadata:
                for metadata in rpc_metadatas.metadata:
                    pretty_metadata += (
                        metadata.key + ": " + metadata.value + ", "
                    )
            result[peer] = pretty_metadata
        return result

    @classmethod
    def from_response(
        cls, lb_stats: grpc_testing.LoadBalancerStatsResponse
    ) -> "PrettyLoadBalancerStats":
        rpcs_by_method: Dict[str, "RpcsByPeer"] = dict()
        for method_name, stats in lb_stats.rpcs_by_method.items():
            if stats:
                rpcs_by_method[method_name] = cls._parse_rpcs_by_peer(
                    stats.rpcs_by_peer
                )
        return PrettyLoadBalancerStats(
            num_failures=lb_stats.num_failures,
            rpcs_by_peer=cls._parse_rpcs_by_peer(lb_stats.rpcs_by_peer),
            rpcs_by_method=rpcs_by_method,
            metadatas_by_peer=cls._parse_metadatas_by_peer(
                lb_stats.metadatas_by_peer
            ),
        )


def lb_stats_pretty(lb: grpc_testing.LoadBalancerStatsResponse) -> str:
    """Pretty print LoadBalancerStatsResponse.

    Example:
      num_failures: 13
      rpcs_by_method:
        UNARY_CALL:
          psm-grpc-server-a: 100
          psm-grpc-server-b: 42
        EMPTY_CALL:
          psm-grpc-server-a: 200
      rpcs_by_peer:
        psm-grpc-server-a: 200
        psm-grpc-server-b: 42
    """
    pretty_lb_stats = PrettyLoadBalancerStats.from_response(lb)
    stats_as_dict = dataclasses.asdict(pretty_lb_stats)

    # Don't print metadatas_by_peer unless it has data
    if not stats_as_dict["metadatas_by_peer"]:
        stats_as_dict.pop("metadatas_by_peer")

    return yaml.dump(stats_as_dict, sort_keys=False)
